import base64
import contextlib
import hashlib
import inspect
import json
import os
import pickle
import subprocess
import sys
import tempfile
from abc import ABC, abstractmethod
from typing import Callable, Union
from unittest.mock import MagicMock

import numpy as np
import pandas as pd
import pyrebase  # Default pyrebase is pyrebase3 which won't work. Need to install pyrebase4 (pip install pyrebase4)
from google.cloud.firestore import Client
from google.oauth2.credentials import Credentials

import olorenchemengine

sys.modules[
    "olorenautoml"
] = olorenchemengine  # important for backwards compatibility of some models


def mock_imports(g, *args):
    for arg in args:
        g[arg] = MagicMock()


def all_subclasses(cls):
    """Helper function to return all subclasses of class"""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


class OASConnector:
    """Class which links oce to OAS and can move data between them using Firestore and an API"""

    def __init__(self) -> None:
        self.storage = None

    def authenticate(self):
        import json

        import requests
        from requests.exceptions import HTTPError

        FIREBASE_REST_API = "https://identitytoolkit.googleapis.com/v1/accounts"
        FIREBASE_API_KEY = "AIzaSyCSaFE5-qxy5-HAr-oqUd5del4bv5dqJ0Q"

        # Use the google verify password REST api to authenticate and generate user tokens
        def log_in_first_time(token):
            request_url = (
                f"{FIREBASE_REST_API}:signInWithCustomToken?key={FIREBASE_API_KEY}"
            )
            headers = {"content-type": "application/json; charset=UTF-8"}
            data = json.dumps({"token": token, "returnSecureToken": True})
            resp = requests.post(request_url, headers=headers, data=data)

            # Check for errors
            try:
                resp.raise_for_status()
            except HTTPError as e:
                raise HTTPError(e, resp.text)

            return resp.json()

        def get_id_token_from_refresh(refresh_token):
            request_url = (
                f"https://securetoken.googleapis.com/v1/token?key={FIREBASE_API_KEY}"
            )
            headers = {"content-type": "application/json; charset=UTF-8"}
            data = json.dumps(
                {"refresh_token": refresh_token, "grant_type": "refresh_token"}
            )
            resp = requests.post(request_url, headers=headers, data=data)
            try:
                resp.raise_for_status()
            except HTTPError as e:
                raise HTTPError(e, resp.text)

            return resp.json()

        from os.path import expanduser

        home = expanduser("~")
        PATH_TO_REFRESH_TOKEN = f"{home}/.oce/firebase_refresh_token.json"

        if not os.path.exists(PATH_TO_REFRESH_TOKEN):
            print("Navigate to https://oas.oloren.ai/auth to get auth token.")
            response = log_in_first_time(input("Paste token here: "))
            if not os.path.exists(f"{home}/.oce"):
                os.makedirs(f"{home}/.oce")
            with open(PATH_TO_REFRESH_TOKEN, "w") as f:
                f.write(json.dumps({"refreshToken": response["refreshToken"]}))

        with open(PATH_TO_REFRESH_TOKEN, "r") as f:
            refresh_token = json.loads(f.read())["refreshToken"]
            response = get_id_token_from_refresh(refresh_token)
            response["idToken"] = response["id_token"]
            response["refreshToken"] = response["refresh_token"]
            response["uid"] = response["user_id"]

        with open(PATH_TO_REFRESH_TOKEN, "w") as f:
            f.write(json.dumps({"refreshToken": response["refreshToken"]}))

        creds = Credentials(response["idToken"], response["refreshToken"])

        """Storage"""

        config = {
            "apiKey": "AIzaSyCSaFE5-qxy5-HAr-oqUd5del4bv5dqJ0Q",
            "authDomain": "oloren-ai.firebaseapp.com",
            "databaseURL": "https://oloren-ai-default-rtdb.firebaseio.com",
            "projectId": "oloren-ai",
            "storageBucket": "oloren-ai.appspot.com",
            "messagingSenderId": "602366687071",
            "appId": "1:602366687071:web:f35531f7142084b86a28ea",
            "measurementId": "G-MVNK4DPQ60",
        }
        self.firebase = pyrebase.initialize_app(config)

        # self.auth = self.firebase.auth()
        # response = self.auth.refresh(response["refreshToken"])

        self.storage = self.firebase.storage()
        self.logging_db = Client("oloren-ai", creds)
        self.uid = response["uid"]
        self.uid_token = response["idToken"]

        return self.uid_token

    def upload_vis(self, visualization):
        self.authenticate()

        out = self.logging_db.collection("visualizations").add(
            {
                "uid": self.uid,
                "status": "Visualization rendered",
                "data_url": visualization.render_data_url(),
            }
        )

        print(f"Uploaded visualization to https://oas.oloren.ai/vis?vid={out[1].id}")

        return out

    def upload_model(self, model, model_name):
        self.authenticate()

        import tempfile

        with tempfile.NamedTemporaryFile() as tmp:
            pickle.dump(saves(model), tmp)
            tmp.flush()
            response = self.storage.child(f"{self.uid}/models/{model_name}.pkl").put(
                tmp.name, self.uid_token
            )

        return self.logging_db.collection("models").add(
            {
                "uid": self.uid,
                "did": None,
                "source": "oce",
                "data": {
                    "file_path": f"gs://oloren-ai.appspot.com/{self.uid}/models/{model_name}.pkl",
                    "name": model_name,
                    "description": pretty_params_str(model),
                    "parameters": json.dumps(parameterize(model)),
                    "metrics": None,
                    "status": "Runnable",
                },
            }
        )


oas_connector = OASConnector()

ignored_kwargs = ["map_location", "num_workers"]


def download_public_file(path, redownload=False):
    """Download a public file from Oloren's Public Storage, and returns the contents.

    @param path: The path to the file to read.
    @param redownload: Whether to redownload the file if it already exists.
    """

    local_path = os.path.join(os.path.expanduser("~"), ".oce", path)

    if os.path.exists(local_path) and not redownload:
        return local_path

    if not os.path.exists(os.path.dirname(local_path)):
        os.makedirs(os.path.dirname(local_path))

    print(f"Downloading {path}...")
    import urllib.request

    urllib.request.urlretrieve(
        f"https://storage.googleapis.com/oloren-public-data/{path}", local_path
    )
    return local_path


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def log_arguments(func: Callable[..., None]) -> Callable[..., None]:
    """
    log_arguments is a decorator which logs the arguments of a BaseClass constructor to instance variables for use in
        model parameterization.

    Args:
        func (function): a __init__(self, *args, **kwargs) function of a baseclass.

    Returns:
        wrapper (function): the same __init__ function with arguments saved to instance variables.
    """

    def wrapper(self, *args, **kwargs):
        if "log" not in kwargs.keys() or kwargs["log"]:
            kwds = get_default_args(func)
            for k, v in kwds.items():
                if k not in kwargs:
                    kwargs[k] = v
            self.args = args
            self.kwargs = {k: v for k, v in kwargs.items() if k not in ignored_kwargs}
        if _runtime.is_local:
            return func(self, *args, **kwargs)
        import uuid

        REMOTE_ID = str(uuid.uuid4())

        if "BaseRemoteSymbol" in str(type(self)):
            self.REMOTE_ID = REMOTE_ID
            return func(self, *args, **kwargs)

        _runtime.add_instruction(
            {
                "type": "CREATE",
                "REMOTE_ID": REMOTE_ID,
                "parameters": parameterize(self),
            }
        )
        self.REMOTE_ID = REMOTE_ID

    wrapper.__wrapped__ = func

    return wrapper


def _create_BC_if_necessary(obj):
    return obj if type(obj) is not dict else create_BC(obj)


def deparametrize_args_kwargs(params):
    args = params["args"]
    kwargs = params["kwargs"]
    return [_create_BC_if_necessary(arg) for arg in args], {
        key: create_BC(kwarg) for key, kwarg in kwargs.items()
    }


def recursive_get_attr(parent, attr):
    if len(attr) == 0:
        return parent
    return recursive_get_attr(getattr(parent, attr[0]), attr[1:])


def generate_uuid():
    import uuid

    return str(uuid.uuid4())


class _RemoteRuntime:
    def __init__(self) -> None:
        self.reset_remote()

    def get_remote_obj(self, remote_id):
        if self.runner is None:
            raise NotImplementedError(
                "Not yet implemented remote object retrieval locally"
            )
        else:
            return self.runner.get_remote_obj(remote_id)

    def add_instruction(self, instruction):

        self.instruction_buffer.append(instruction)

        if instruction["type"] == "CALL":
            x = self.send_instructions_blocking()
            if x is not None and isinstance(x, str):
                return json.loads(x)

    def get_iterable(self, remote_id):
        self.instruction_buffer.append({"type": "ITER", "REMOTE_ID": remote_id})
        return [
            BaseRemoteSymbol.from_rid(rid) for rid in self.send_instructions_blocking()
        ]

    def get_obj_repr(self, remote_id):
        self.instruction_buffer.append({"type": "REPR", "REMOTE_ID": remote_id})
        return self.send_instructions_blocking()

    def send_instructions_blocking(self):
        import requests

        response = None

        oas_connector.authenticate()
        _, execution = oas_connector.logging_db.collection("executions").add(
            {
                "uid": oas_connector.uid,
                "sid": self.session_id,
                "instructions": json.dumps(self.instruction_buffer),
                "status": "Queued",
            }
        )

        eid = execution.id

        response = requests.post(
            f"{self.remote_url}/firestore/run_remote/",
            params={
                "eid": eid,
                "uid": oas_connector.authenticate(),
            },
            headers={
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
            },
        )

        if response.status_code != 200:
            self.instruction_buffer = []
            raise ValueError(f"Error code on post to {self.remote_url} {response.status_code} - {response.text}. Length of instruction buffer: {len(json.dumps(self.instruction_buffer))}")

        response = response.json()

        execution = None

        def on_snapshot(doc_snapshot, changes, read_time):
            nonlocal execution
            for doc in doc_snapshot:
                execution = doc.to_dict()

        listener = (
            oas_connector.logging_db.collection("executions")
            .document(eid)
            .on_snapshot(on_snapshot)
        )

        while execution is None:
            pass

        while execution["status"] not in ("Finished", "Error"):
            pass

        listener.close()

        if execution["status"] == "Error":
            print(f"Remote Call: {self.instruction_buffer}\nResulted in Traceback:")
            print(execution["traceback"])
            oas_connector.logging_db.collection("executions").document(eid).delete()
            self.instruction_buffer = []
            raise ValueError("Remote call resulted in error")

        elif execution["status"] == "Finished":
            if len(execution["stdout"].strip()) > 0:
                print(
                    "\n".join([f"REMOTE: {x}" for x in execution["stdout"].split("\n")])
                )
            oas_connector.logging_db.collection("executions").document(eid).delete()
            self.instruction_buffer = []
            return execution["return"]
        else:
            raise NotImplementedError(f"Unknown status {execution['status']}")

    @property
    def is_local(self):
        return self.runtime == "local"

    def start_remote(self, session_id, debug, remote_url):
        self.runtime = "remote"
        self.instruction_buffer = []
        self.session_id = session_id
        self.runner = None
        self.debug = debug
        self.remote_url = remote_url

    def reset_remote(self):
        self.runtime = "local"
        self.instruction_buffer = []
        self.session_id = None
        self.runner = None
        self.debug = False
        self.remote_url = None


_runtime = _RemoteRuntime()  # internal runtime object


def set_runner(runner):
    _runtime.runner = runner


def get_runtime():
    return _runtime


class Remote(object):
    def __init__(self, remote_url, session_id=None, keep_alive=False, debug=False):
        self.keep_alive = keep_alive
        self.session_id = session_id
        self.debug = debug
        self.remote_url = remote_url

    def __enter__(self):
        self.session_id = generate_uuid()
        _runtime.start_remote(self.session_id, self.debug, self.remote_url)
        return self.session_id

    def __exit__(self, type, value, traceback):
        _runtime.reset_remote()
        if not self.keep_alive:
            import google

            oas_connector.authenticate()

            with contextlib.suppress(google.api_core.exceptions.PermissionDenied):
                doc = (
                    oas_connector.logging_db.collection("sessions")
                    .document(self.session_id)
                    .get()
                    .to_dict()
                )
                if doc is not None:
                    for rid in doc["objects"]:
                        try:
                            oas_connector.storage.delete(
                                f"{oas_connector.uid}/sessions/{self.session_id}/{rid}.oce",
                                oas_connector.uid_token,
                            )
                        except:
                            import logging

                            logging.warning(
                                f"Unable to delete remote object {rid} from session {self.session_id}"
                            )
                    oas_connector.logging_db.collection("sessions").document(
                        self.session_id
                    ).delete()


def pretty_args_kwargs(args, kwargs):
    parameterized_args = [parameterize(arg) for arg in args]
    parameterized_kwargs = {k: parameterize(v) for k, v in kwargs.items()}
    return f"(args={parameterized_args}, kwargs={parameterized_kwargs})"


def _paremetrize_if_necessary(obj):
    if isinstance(obj, BaseRemoteSymbol):
        return parameterize(obj)
    elif isinstance(obj, pd.Series):
        return list(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    with contextlib.suppress(Exception):
        json.dumps(obj)
        return obj


def parametrize_args_kwargs(args, kwargs):
    parameterized_args = [_paremetrize_if_necessary(arg) for arg in args]
    parameterized_kwargs = {k: _paremetrize_if_necessary(v) for k, v in kwargs.items()}
    return {"args": parameterized_args, "kwargs": parameterized_kwargs}


def _truncate_json(json_obj, max_length=40):
    if isinstance(json_obj, str):
        if len(json_obj) > max_length:
            return f"{json_obj[:max_length]}..."
        else:
            return json_obj
    elif isinstance(json_obj, list):
        return [_truncate_json(x, max_length) for x in json_obj]
    elif isinstance(json_obj, dict):
        return {
            _truncate_json(k): _truncate_json(v, max_length)
            for k, v in json_obj.items()
        }
    else:
        return json_obj


class BaseRemoteSymbol:
    @log_arguments
    def __init__(
        self, REMOTE_SYMBOL_NAME, REMOTE_PARENT, args=None, kwargs=None
    ) -> None:
        if _runtime.is_local:
            raise RuntimeError("Cannot instantiate RemoteObject in local _runtime")

        if args is not None:
            self.args = args
        if kwargs is not None:
            self.kwargs = kwargs

        self.REMOTE_SYMBOL_NAME = REMOTE_SYMBOL_NAME
        self.REMOTE_PARENT = REMOTE_PARENT
        self.REMOTE_CHILDREN = {}

        if REMOTE_SYMBOL_NAME == "CALL":
            out = _runtime.add_instruction(
                {
                    "type": "CALL",
                    "PARENT_REMOTE_ID": self.REMOTE_PARENT.REMOTE_ID,
                    "REMOTE_ID": self.REMOTE_ID,
                    "ARGUMENTS": parametrize_args_kwargs(args, kwargs),
                }
            )

            if self.REMOTE_PARENT.REMOTE_SYMBOL_NAME == "render_ipynb":
                from IPython.display import IFrame, display

                display(IFrame(out, width=800, height=600))

        else:
            if not hasattr(self.REMOTE_PARENT, "REMOTE_ID"):
                self.REMOTE_PARENT._upload_remote()

            _runtime.add_instruction(
                {
                    "type": "SYMBOL",
                    "PARENT_REMOTE_ID": self.REMOTE_PARENT.REMOTE_ID,
                    "SYMBOL_NAME": self.REMOTE_SYMBOL_NAME,
                    "REMOTE_ID": self.REMOTE_ID,
                }
            )

    @classmethod
    def from_rid(cls, rid):
        x = object.__new__(cls)
        x.REMOTE_ID = rid
        return x

    def __iter__(self):
        return iter(_runtime.get_iterable(self.REMOTE_ID))

    def __repr__(self):
        return _runtime.get_obj_repr(self.REMOTE_ID)

    def _upload_remote(self):
        try:
            oas_connector.authenticate()
            REMOTE_ID = generate_uuid()

            r = _runtime.runtime
            _runtime.runtime = "local"

            with tempfile.NamedTemporaryFile() as tmp:
                pickle.dump(saves(self), tmp)
                tmp.flush()
                response = oas_connector.storage.child(
                    f"{oas_connector.uid}/sessions/{_runtime.session_id}"
                    + f"/{REMOTE_ID}.oce"
                ).put(tmp.name, oas_connector.uid_token)
        except:
            import sys
            import traceback

            traceback.print_exc()
            sys.exit(0)

        _runtime.runtime = r
        object.__setattr__(self, "REMOTE_ID", REMOTE_ID)

        return REMOTE_ID

    def __getattribute__(self, key):
        if _runtime.is_local or key == "_upload_remote":
            return object.__getattribute__(self, key)
        if "ipython_canary_method_should_not_exist" in key:
            return {}
        if (
            key.startswith("REMOTE")
            or key.startswith("__")
            or key in ["args", "kwargs"]
        ):
            return object.__getattribute__(self, key)

        if not hasattr(self, "REMOTE_CHILDREN") or key not in self.REMOTE_CHILDREN:
            if not hasattr(self, "REMOTE_CHILDREN"):
                self.REMOTE_CHILDREN = {}
            self.REMOTE_CHILDREN[key] = BaseRemoteSymbol(key, self)
        return self.REMOTE_CHILDREN[key]

    def __call__(self, *args, **kwargs):
        # TODO: move logic from init to here, then return baseremotesymbol only if return type is none

        remote_id = generate_uuid()
        out = _runtime.add_instruction(
            {
                "type": "CALL",
                "PARENT_REMOTE_ID": self.REMOTE_ID,
                "REMOTE_ID": remote_id,
                "ARGUMENTS": parametrize_args_kwargs(args, kwargs),
            }
        )

        if self.REMOTE_SYMBOL_NAME == "render_ipynb":
            from IPython.display import IFrame, display

            display(IFrame(out, width=800, height=600))

        return out if out is not None else RemoteObj(remote_id)


class BaseClass(BaseRemoteSymbol):
    """BaseClass is the base class for all models.

    All classes in Oloren ChemEngine should inherit from BaseClass to enable for universal saving and loading of both
    parameters and internal state. This requires the implementation of abstract methods _save and _load.

    Methods:
        Registry: returns a dictionary mapping the name of a class to the class itself for all subclasses of the class.
        _save: saves an instance of a BaseClass to a dictionary (abstract method to be implmented by subclasses)
        _load: loads an instance of a BaseClass from a dictionary (abstract method to be implmented by subclasses)
    """

    def __init__(self):
        pass

    @classmethod
    def Registry(cls):
        """Registry is a recursive method to create a dictionary of all subclasses of BaseClass, with the key being the name of the subclass and the value being the subclass itself."""
        d = {cls.__name__: cls}
        scs = cls.__subclasses__()
        if len(scs) > 0:
            for sc in scs:
                u = sc.Registry()
                d = {**d, **u}
        return d

    @classmethod
    def AllInstances(cls):
        """AllTypes returns a list of all standard instances of all subclasses of BaseClass.

        Standard instances means that all required parameters for instantiation of the
        subclasses are set with canonical values."""
        if hasattr(cls, "__abstractmethods__") and len(cls.__abstractmethods__) != 0:
            return [o for sc in cls.__subclasses__() for o in sc.AllInstances()]
        try:
            return [cls()] + [
                o for sc in cls.__subclasses__() for o in sc.AllInstances()
            ]
        except Exception:
            return [o for sc in cls.__subclasses__() for o in sc.AllInstances()]

    @classmethod
    def Opt(cls, *args, **kwargs):
        return {
            **{"BC_class_name": cls.__name__},
            **{"args": args},
            **{"kwargs": kwargs},
        }

    @abstractmethod
    def _save(self) -> dict:
        """_save is an abstract method that must be implemented by all subclasses. It should return a dictionary of variables which can passed to an instance of a model via _load to completely recreate a model.

        Returns:
            dict: a dictionary of variables which can be passed to _load to recreate the model.
        """
        return ""

    @abstractmethod
    def _load(self, d: dict):
        """_load is an abstract method that must be implemented by all subclasses. It should take an instance save by _save and recreate the model."""
        pass

    def copy(self):
        obj_copy = create_BC(parameterize(self))
        obj_copy._load(self._save())
        return obj_copy


class RemoteObj(BaseRemoteSymbol):
    """Dummy object to represent remote objects."""

    def __init__(self, remote_id):
        self.REMOTE_ID = remote_id


def parameterize(object: Union[BaseClass, list, int, float, str, None]) -> dict:
    """parameterize is a recursive method which creates a dictionary of all arguments necessary to instantiate a BaseClass object.

    Note that only objects which are instances of subclasses of BaseClass can be parameterized, other supported objects are to enable to recursive use of parameterize but cannot themselves be parameterized.

    Args:
        object (Union[BaseClass, list, int, float, str, None]): parameterize is a recursive method which creates a dictionary of all arguments necessary to instantiate a BaseClass object.

    Raises:
        ValueError: Object is not of type that can be parameterized

    Returns:
        dict: dictionary of parameters necessary to instantiate the object.
    """
    if issubclass(type(object), BaseClass) or (
        issubclass(type(object), BaseRemoteSymbol) and hasattr(object, "REMOTE_ID")
    ):
        if hasattr(object, "REMOTE_ID"):
            return {"REMOTE_ID": object.REMOTE_ID}
        return {
            **{"BC_class_name": type(object).__name__},
            **{"args": [parameterize(arg) for arg in object.args]},
            **{"kwargs": {k: parameterize(v) for k, v in object.kwargs.items()}},
        }
    elif (
        object is None
        or issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
    ):
        return object
    elif issubclass(type(object), list):
        return [parameterize(x) for x in object]
    else:
        raise ValueError(f"Invalid object {object}")


def model_name_from_params(param_dict: dict) -> str:
    """model_name_from_params creates a unique name for a model based on the parameters passed to it.

    Args:
        param_dict (dict): dictionary of parameters returned by parameterize neccessary to instantiate the model (note this is different from the instance save)

    Returns:
        str: the model name consisting of the the model class name with a hash of the parameters
    """
    return (
        param_dict["BC_class_name"]
        + " "
        + base64.urlsafe_b64encode(
            hashlib.md5(str(param_dict).encode("utf-8")).digest()
        )[:8].decode("utf-8")
    )


def model_name_from_model(model: BaseClass) -> str:
    """model_name_from_model creates a unique name for a model.

    Args:
        model (BaseClass): the model to be named

    Returns:
        str: the model name consisting of the the model class name with a hash of the parameters
    """
    return model_name_from_params(parameterize(model))


def saves(object: Union[BaseClass, dict, list, int, float, str, None]) -> dict:
    """saves is a method which saves BaseClass object, which can be recovered via loads.

    Args:
        object (Union[BaseClass, dict, list, int, float, str, None]): the object to be saved

    Returns:
        dict: a dictionary which can be passed to loads to recreate the object
    """
    if issubclass(type(object), BaseClass):
        return {
            **{"BC_class_name": type(object).__name__},
            **{"instance_save": object._save()},
            **{"args": [saves(arg) for arg in object.args]},
            **{"kwargs": {k: saves(v) for k, v in object.kwargs.items()}},
        }
    elif (
        issubclass(type(object), int)
        or issubclass(type(object), float)
        or issubclass(type(object), str)
        or object is None
    ):
        return object
    elif issubclass(type(object), list):
        return [saves(x) for x in object]
    elif issubclass(type(object), dict):
        return {k: saves(v) for k, v in object.items()}
    else:
        print(object)
        raise ValueError


def create_BC(d: dict) -> BaseClass:
    """create_BC is a method which creates a BaseClass object from a dictionary of parameters.

    Note the instances variables of the object are not specified.

    Args:
        d (dict): a dictionary of parameters returned by parameterize

    Returns:
        BaseClass: the object created from the parameters
    """
    if isinstance(d, str):
        d = (
            d.replace("'", '"')
            .replace("True", "true")
            .replace("False", "false")
            .replace("None", "null")
        )
        d = json.loads(d)

    if "REMOTE_ID" in d.keys():
        return _runtime.get_remote_obj(d["REMOTE_ID"])

    args = []
    if "args" in d.keys():
        for arg in d["args"]:
            if isinstance(arg, dict) and (
                "BC_class_name" in arg.keys() or "REMOTE_ID" in arg.keys()
            ):
                args.append(create_BC(arg))
            else:
                if isinstance(arg, list):
                    arg = [
                        create_BC(x)
                        if isinstance(x, dict)
                        and ("BC_class_name" in x.keys() or "REMOTE_ID" in x.keys())
                        else x
                        for x in arg
                    ]
                args.append(arg)

    kwargs = {}
    if "kwargs" in d.keys():
        for k, v in d["kwargs"].items():
            if isinstance(v, dict) and (
                "BC_class_name" in v.keys() or "REMOTE_ID" in v.keys()
            ):
                kwargs[k] = create_BC(v)
            else:
                kwargs[k] = v

    return BaseClass.Registry()[d["BC_class_name"]](*args, **kwargs)


def loads(d: dict) -> BaseClass:
    """loads is a method which recreates a BaseClass object from a save.

    Args:
        d (dict): the dictionary returned by saves which saves the state of a BaseClass object

    Returns:
        BaseClass: the recreated object
    """

    import olorenchemengine as olorenautoml  # for backwards compatibility

    args = []
    for arg in d["args"]:
        if isinstance(arg, dict) and "BC_class_name" in arg.keys():
            args.append(loads(arg))
        else:
            if isinstance(arg, list):
                arg = [
                    loads(x)
                    if isinstance(x, dict) and "BC_class_name" in x.keys()
                    else x
                    for x in arg
                ]
            args.append(arg)

    kwargs = {
        k: loads(v) if isinstance(v, dict) and "BC_class_name" in v.keys() else v
        for k, v in d["kwargs"].items()
    }
    bc = BaseClass.Registry()[d["BC_class_name"]](*args, **kwargs)
    bc._load(d["instance_save"])
    return bc


def save(model: BaseClass, fname: str):
    """saves a BaseClass object to a file

    Args:
        model (BaseClass): the object to be saved
        fname (str): the file name to save the model to
    """

    if hasattr(model, "REMOTE_ID"):

        REMOTE_ID = model.REMOTE_ID
        oas_connector.storage.child(
            f"{oas_connector.uid}/sessions/{_runtime.session_id}/{REMOTE_ID}.oce.delete"
        ).download("", fname, token=oas_connector.authenticate())
    else:
        save_dict = saves(model)
        with open(fname, "wb+") as f:
            pickle.dump(save_dict, f)


def load(fname: str) -> BaseClass:
    """loads a BaseClass from a file

    Args:
        fname (str): name of the file to load the object from

    Returns:
        BaseClass: the BaseClass object which as been recreated from the file
    """
    if _runtime.is_local:
        with open(fname, "rb") as f:
            d = pickle.load(f)
        return loads(d)
    else:
        oas_connector.authenticate()
        REMOTE_ID = generate_uuid()
        response = oas_connector.storage.child(
            f"{oas_connector.uid}/sessions/{_runtime.session_id}" + f"/{REMOTE_ID}.oce.delete"
        ).put(fname, oas_connector.uid_token)
        return RemoteObj(REMOTE_ID)


def pretty_params(base: Union[BaseClass, dict]) -> dict:
    """Returns a dictionary of the parameters of the passed BaseClass object, formatted such that they are in a
    human readable format, with the names of the arguments included."""
    if isinstance(base, dict):
        base = loads(base)

    if issubclass(type(base), BaseClass):
        args = list(base.args)
        kwargs = dict(base.kwargs.items())
        base_object_parameters = list(
            inspect.signature(base.__init__).parameters.keys()
        )

        for kwarg in kwargs:
            if kwarg in base_object_parameters:
                base_object_parameters.remove(kwarg)

        labeled_args = dict(zip(base_object_parameters, args))
        fully_labeled_args = {k: pretty_params(v) for k, v in labeled_args.items()}
        fully_labeled_kwargs = {k: pretty_params(v) for k, v in kwargs.items()}

        return {
            **{"BC_Class_name": type(base).__name__},
            **fully_labeled_args,
            **fully_labeled_kwargs,
        }
    elif (
        issubclass(type(base), int)
        or issubclass(type(base), float)
        or issubclass(type(base), str)
        or base is None
    ):
        return base
    elif issubclass(type(base), list):
        return [pretty_params(x) for x in base]
    elif issubclass(type(base), dict):
        return {k: pretty_params(v) for k, v in base.items()}
    else:
        print(type(base))
        raise ValueError


def pretty_params_str(base: Union[BaseClass, dict]) -> str:
    """Returns a string of the parameters of the passed BaseClass object, formatted such that they are in a human
    readable format"""
    return json.dumps(pretty_params(base), indent=4)


def json_params_str(base: Union[BaseClass, dict]) -> str:
    """Returns a json string of the parameters of the passed BaseClass object so that the model parameter dictionary can
    be reconstructed with json.load(params_str)"""
    return (
        json.dumps(pretty_params(base))
        .replace("'", '"')
        .replace("True", "true")
        .replace("False", "false")
        .replace("None", "null")
    )


def install_with_permission(package_name: str):
    inp = input(
        f"The required package {package_name} is not installed. Do you want to install it? [y/N]? "
    )
    if inp.lower() == "y":
        subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])
    else:
        print(
            f"Stopping program. You can install the package manually with: \n >> pip install {package_name}"
        )
        os._exit(1)


def import_or_install(package_name: str, statement: str = None, scope: dict = None):
    if scope is None:
        scope = globals()
    if statement is None:
        statement = f"import {package_name}"
    try:
        exec(statement, scope)
    except ImportError:
        install_with_permission(package_name)
    finally:
        exec(statement, scope)



from cmath import exp
from tkinter import N

import numpy as np
import pandas as pd
from rdkit.Chem import AllChem
from rdkit.DataStructs.cDataStructs import BulkTanimotoSimilarity
from sklearn.cross_decomposition import PLSRegression
from sklearn.decomposition import PCA
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import OneHotEncoder

import olorenchemengine as oce

from .base_class import *
from .basics import *
from .dataset import *
from .representations import *


class BaseFingerprintModel(BaseErrorModel):
    """BaseFingerprintModel is the base class for error models that require the
    computation of Morgan Fingerprints.
    """

    def build(
        self,
        model: BaseModel,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y: Union[np.ndarray, list, pd.Series],
    ):
        super().build(model, X, y)
        if isinstance(self.X_train, pd.DataFrame):
            self.X_train.columns = ["SMILES"]
        smiles = SMILESRepresentation().convert(self.X_train)
        mols = [Chem.MolFromSmiles(smi) for smi in smiles]
        self.train_fps = [AllChem.GetMorganFingerprint(mol, 2) for mol in mols]

    def _save(self) -> dict:
        d = super()._save()
        if hasattr(self, "train_fps"):
            d.update({"train_fps": self.train_fps})
        return d

    def _load(self, d) -> None:
        super()._load(d)
        if "train_fps" in d.keys():
            self.train_fps = d["train_fps"]


class SDC(BaseFingerprintModel):
    """SDC is an error model that predicts error bars based on the Sum of
        Distance-weighted Contributions: `Molecular Similarity-Based Domain
        Applicability Metric Efficiently Identifies Out-of-Domain Compounds
        <http://dx.doi.org/10.1021/acs.jcim.8b00597>`_

        Parameters:
            a (int or float, optional): Value of a in the SDC formula. Default 3.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.SDC())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, a=3):
        self.a = a

    def calculate(self, X, y_pred):
        def sdc(smi):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            TD = 1 - np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            return np.sum(np.exp(-self.a * TD / (1 - TD)))

        return np.array([sdc(smi) for _, smi in tqdm(np.ndenumerate(X))])


class TargetDistDC(SDC):
    """TargetDistDC is an error model that calculates the root-mean-square
        difference between the predicted activity of the target molecule and
        the observed activities of all training molecules, weighted by the DC.

        Parameters:
            a (int or float, optional): Value of a in the SDC formula. Default 3.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TargetDistDC())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    def calculate(self, X, y_pred):
        def dist(smi, pred):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            TD = 1 - np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            DC = np.exp(-self.a * TD / (1 - TD))
            return np.sqrt(np.dot(DC, np.abs(self.y_train - pred)) / np.sum(DC))

        X = np.array(X).flatten()
        return np.array([dist(smi, y_pred[i]) for i, smi in tqdm(np.ndenumerate(X))])


class TrainDistDC(SDC):
    """TrainDistDC is an error model that calculates the root-mean-square
        difference between the predicted and observed activities of all
        training molecules, weighted by the DC.

        Parameters:
            a (int or float, optional): Value of a in the SDC formula. Default 3.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TrainDistDC())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    def calculate(self, X, y_pred):
        residuals = np.abs(self.y_train - self.y_pred_train)

        def dist(smi):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            TD = 1 - np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            DC = np.exp(-self.a * TD / (1 - TD))
            return np.sqrt(np.dot(DC, residuals) / np.sum(DC))

        return np.array([dist(smi) for _, smi in tqdm(np.ndenumerate(X))])


class KNNSimilarity(BaseFingerprintModel):
    """NNSimilarity is an error model that calculates mean Tanimoto similarity
        between the target molecule and the k most similar training molecules
        using a Morgan Fingerprint with a radius of 2 bonds.

        Parameters:
            k (int, optional): Number of nearest neighbors to consider. Default 5.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.KNNSimilarity())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, k=5):
        self.k = k

    def calculate(self, X, y_pred):
        def mean_sim(smi):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            similarity = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            return np.mean(sorted(similarity)[-self.k :])

        return np.array([mean_sim(smi) for _, smi in tqdm(np.ndenumerate(X))])


class TargetDistKNN(KNNSimilarity):
    """TargetDistKNN is an error model that calculates the root-mean-square
        difference between the predicted activity of the target molecule and
        the observed activities of the k most similar training molecules,
        weighted by their similarity.

        Parameters:
            k (int, optional): Number of nearest neighbors to consider. Default 5.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TargetDistKNN())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    def calculate(self, X, y_pred):
        def dist(smi, pred):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            similarity = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            mae = np.abs(self.y_train - pred)
            idxs = np.argsort(similarity)[-self.k :]
            return np.sqrt(
                np.dot(similarity[idxs], mae[idxs]) / np.sum(similarity[idxs])
            )

        X = np.array(X).flatten()
        return np.array([dist(smi, y_pred[i]) for i, smi in tqdm(np.ndenumerate(X))])


class TrainDistKNN(KNNSimilarity):
    """TrainDistKNN is an error model that calculates the root-mean-square
        difference between the predicted and observed activities of the k most
        similar training molecules to the target molecule, weighted by their
        similarity.

        Parameters:
            k (int, optional): Number of nearest neighbors to consider. Default 5.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.TrainDistKNN())
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    def calculate(self, X, y_pred):
        residuals = np.abs(self.y_train - self.y_pred_train)

        def dist(smi):
            mol = Chem.MolFromSmiles(smi)
            ref_fp = AllChem.GetMorganFingerprint(mol, 2)
            similarity = np.array(BulkTanimotoSimilarity(ref_fp, self.train_fps))
            idxs = np.argsort(similarity)[-self.k :]
            return np.sqrt(
                np.dot(similarity[idxs], residuals[idxs]) / np.sum(similarity[idxs])
            )

        return np.array([dist(smi) for _, smi in tqdm(np.ndenumerate(X))])


class Predicted(BaseErrorModel):
    """Predicted is an error model that predicts error bars based on only the
        predicted value of a molecule. It is best used as part of an aggregate
        error model rather than by itself.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.RandomForestErrorModel([oce.SDC(), oce.Predicted()])
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self):
        pass

    def calculate(self, X, y_pred):
        return y_pred


class ADAN(BaseErrorModel):
    """ADAN is an error model that predicts error bars based on one or
        multiple ADAN categories: `Applicability Domain Analysis (ADAN): A
        Robust Method for Assessing the Reliability of Drug Property Predictions
        <https://doi.org/10.1021/ci500172z>`_

        Parameters:
            criteria (Union[str, list, set]): the ADAN criteria to be considered

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.ADAN("E_raw"))
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    @log_arguments
    def __init__(self, criterion: str):
        self.criterion = criterion

    def build(
        self,
        model: BaseModel,
        X: Union[pd.DataFrame, np.ndarray, list, pd.Series],
        y: Union[np.ndarray, list, pd.Series],
        rep: BaseCompoundVecRepresentation = None,
        dim_reduction: str = "pls",
        explvar: float = 0.8,
        threshold: float = 0.95,
    ):
        super().build(model, X, y)
        self.rep = rep

        self.X_train = self.preprocess(self.X_train)
        thresh = int(self.X_train.shape[0] * threshold)
        max_components = min(100, min(self.X_train.shape))

        if dim_reduction == "pls":
            self.reduction = PLSRegression(n_components=max_components)
            self.reduction.fit(self.X_train, self.y_train)
            x_var = np.var(self.reduction.x_scores_, axis=0)
            x_var /= np.sum(x_var)
        elif dim_reduction == "pca":
            self.reduction = PCA(n_components=max_components)
            self.reduction.fit(self.X_train, self.y_train)
            x_var = self.reduction.explained_variance_ratio_
        else:
            raise NameError("dim_reduction {} is not recognized".format(dim_reduction))

        self.dim_reduction = dim_reduction

        if np.sum(x_var) > explvar:
            self.n_components = np.where(np.cumsum(x_var) >= explvar)[0][0] + 1
        else:
            self.n_components = max_components

        self.Xp_train = self.reduction.transform(self.X_train)[:, : self.n_components]
        self.Xp_mean = np.mean(self.Xp_train, axis=0)
        self.y_mean = np.mean(self.y_train)

        nbrs = NearestNeighbors(n_neighbors=2).fit(self.Xp_train)
        distances, indices = nbrs.kneighbors(self.Xp_train)

        centroid_dist = np.linalg.norm(self.Xp_train - self.Xp_mean, axis=1)
        neighbor_dist = distances[:, 1]
        model_dist = self.DModX(self.X_train, self.Xp_train)
        y_mean_dist = np.abs(self.y_train - self.y_mean)
        y_nei_dist = np.abs(self.y_train - self.y_train[indices[:, 1]])
        SDEP_dist = self.SDEP(self.Xp_train, n_drop=1)

        self.training_ = {
            "A_raw": centroid_dist,
            "B_raw": neighbor_dist,
            "C_raw": model_dist,
            "D_raw": y_mean_dist,
            "E_raw": y_nei_dist,
            "F_raw": SDEP_dist,
        }

        self.thresholds_ = {
            "A": sorted(centroid_dist)[thresh],
            "B": sorted(neighbor_dist)[thresh],
            "C": sorted(model_dist)[thresh],
            "D": sorted(y_mean_dist)[thresh],
            "E": sorted(y_nei_dist)[thresh],
            "F": sorted(SDEP_dist)[thresh],
        }

    def calculate_full(self, X):
        criteria = ["A", "B", "C", "D", "E", "F"]
        y_pred = np.array(self.model.predict(X)).flatten()
        X = self.preprocess(X)
        Xp = self.reduction.transform(X)[:, : self.n_components]

        self.results = {}
        for c in criteria:
            self.results[c] = self._calculate(X, Xp, y_pred, c)
        for c in criteria:
            c += "_raw"
            self.results[c] = self._calculate(X, Xp, y_pred, c)
        self.results["Category"] = np.sum([self.results[c] for c in criteria], axis=0)

        self.results = pd.DataFrame(self.results)

    def calculate(self, X, y_pred):
        """Calcualtes confidence scores."""
        X = self.preprocess(X)
        Xp = self.reduction.transform(X)[:, : self.n_components]

        return self._calculate(X, Xp, y_pred, self.criterion)

    def _calculate(self, X, Xp, y_pred, criterion: str, standardize: bool = True):
        if criterion in ("A", "A_raw"):
            dist = np.linalg.norm(Xp - self.Xp_mean, axis=1)
        elif criterion in ("B", "B_raw"):
            nbrs = NearestNeighbors(n_neighbors=1).fit(self.Xp_train)
            distances = nbrs.kneighbors(Xp)[0]
            dist = distances.flatten()
        elif criterion in ("C", "C_raw"):
            dist = self.DModX(X, Xp)
        elif criterion in ("D", "D_raw"):
            dist = self.SDEP(Xp, 0)
        elif criterion in ("E", "E_raw"):
            dist = np.abs(y_pred - self.y_mean)
        elif criterion in ("F", "F_raw"):
            nbrs = NearestNeighbors(n_neighbors=1).fit(self.Xp_train)
            indices = nbrs.kneighbors(Xp)[1]
            dist = np.abs(y_pred - self.y_train[indices.flatten()])
        else:
            raise NameError("criterion {} is not recognized".format(criterion))

        if "raw" in criterion:
            if standardize:
                return (dist - np.mean(self.training_[criterion])) / np.std(
                    self.training_[criterion]
                )
            else:
                return dist
        else:
            return np.where(dist > self.thresholds_[criterion], 1, 0)

    def preprocess(self, X, y=None):
        """Preprocesses data into the appropriate representation."""
        if self.rep is None:
            X = self.model.preprocess(X, y)
        else:
            X = np.array(self.rep.convert(X))

        assert isinstance(
            X, np.ndarray
        ), "The preprocess for the model must return a np.ndarray, e.g. the model must use a BaseCompoundVecRepresentation or a representation must be passed"

        return X

    def DModX(self, X: np.ndarray, Xp: np.ndarray) -> np.ndarray:
        """Computes the distance to the model (DmodX).

        Computes the distance between a datapoint and the PLS model plane. See
        <https://www.jmp.com/support/help/en/15.2/index.shtml#page/jmp/dmodx-calculation.shtml>
        for more details about the statistic.

        Parameters:
            X (np.ndarray): queries
            Xp (np.ndarray): queries transformed into latent space
        """
        if self.dim_reduction == "pls":
            X_reconstructed = (
                Xp
                @ self.reduction.x_loadings_[:, : self.n_components].T
                * self.reduction._x_std
                + self.reduction._x_mean
            )
        elif self.dim_reduction == "pca":
            X_reconstructed = (
                np.dot(Xp, self.reduction.components_[: self.n_components, :])
                + self.reduction.mean_
            )
        return np.linalg.norm(X - X_reconstructed, axis=1) / np.sqrt(
            self.X_train.shape[1] - self.n_components
        )

    def SDEP(
        self, Xp: np.ndarray, n_drop: int = 0, neighbor_thresh: float = 0.05
    ) -> np.ndarray:
        """Computes the standard deviation error of predictions (SDEP).

        Computes the standard deviation training error of the `neighbor_thresh`
        fraction of closest training queries to each query in `Xp` in latent space.

        Parameters:
            Xp (np.ndarray): queries transformed into latent space
            n_drop (int): 1 if fitting, 0 if scoring
            neighbor_thresh (float): fraction of closest training queries to consider
        """
        n_neighbors = int(self.X_train.shape[0] * neighbor_thresh) + n_drop
        nbrs = NearestNeighbors(n_neighbors=n_neighbors).fit(self.Xp_train)
        distances, indices = nbrs.kneighbors(Xp)

        y_sqerr = np.array((self.y_train - self.y_pred_train) ** 2)
        y_mse = np.mean(y_sqerr[indices[:, n_drop:]], axis=1).astype(float)

        return np.sqrt(y_mse).flatten()


class RandomForestErrorModel(BaseAggregateErrorModel):
    """RandomForestErrorModel is an aggregate error model that predicts error
        bars based on an aggregate score determined by a random forest model
        trained on several BaseErrorModels.

        Parameters:
            error_models (Union[BaseErrorModel, List[BaseErrorModel]]): a list
                of error models whose scores will be features for the random
                forest model

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.RandomForestErrorModel([oce.SDC(), oce.KNNSimilarity()])
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    def _train(self, scores, residuals):
        self.aggregate_model = RandomForestRegressor(**self.kwargs)
        self.aggregate_model.fit(scores, residuals)

    def _predict(self, scores, **kwargs):
        return self.aggregate_model.predict(scores, **kwargs)


class LinearRegressionErrorModel(BaseAggregateErrorModel):
    """LinearRegressionErrorModel is an aggregate error model that predicts error
        bars based on a linear combination score from several BaseErrorModels.

        Parameters:
            error_models (Union[BaseErrorModel, List[BaseErrorModel]]): a list
                of error models whose scores will be features for the random
                forest model

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    model.fit_cv(train["Drug"], train["Y"], error_model = oce.LinearRegressionErrorModel([oce.SDC(), oce.KNNSimilarity()])
    model.predict(test["Drug"], return_ci = True)
    ------------------------------
    """

    def _train(self, scores, residuals):
        self.aggregate_model = LinearRegression(**self.kwargs)
        self.aggregate_model.fit(scores, residuals)

    def _predict(self, scores, **kwargs):
        return self.aggregate_model.predict(scores, **kwargs)


class BaseDomainApplicability(BaseClass):
    """Depricated. See BaseErrorModel.

    Base class for analyzing domain applicability of models and estimating
    prediction uncertainty.

    Parameters:
        model (BaseModel): trained model
        rep (BaseCompoundVecRepresentation): preprocessing representation

    Methods:
        _fit: fits a method to a trained model and preprocessed training dataset, to be implemented internally by child classes
        fit: fits a method to a trained model and training dataset
        _test evaluates prediction uncertainty on a preprocessed testing dataset, to be implemented internally by child classes
        test: evaluates prediction uncertainty on a testing dataset
    """

    @log_arguments
    def __init__(self, model: BaseModel, rep: BaseCompoundVecRepresentation = None):
        self.model = model
        self.rep = rep
        self.results = None

    @abstractmethod
    def _fit(self, X, y, y_pred, **kwargs):
        """To be implemented by the child class; fits the method on the internally stored training dataset.

        Note: _fit shouldn"t be directly called by the user, rather it should be called indirectly via the fit method.

        Parameters:
            X (np.ndarray): features, SMILES
            y (np.ndarray): values
            y_pred (np.ndarray): predicted values
        """
        pass

    def fit(self, dataset: BaseDataset, use_entire_dataset: bool = False, **kwargs):
        """Fits model uncertainty to a training dataset. Calls the _fit method on the provided dataset.

        Parameters:
            dataset (BaseDataset): dataset used for fitting the method
            use_entire_dataset (bool): whether to use the entire dataset or only the training data
        """
        if use_entire_dataset:
            X = dataset.entire_dataset[0]
            y = np.array(dataset.entire_dataset[1])
        else:
            X = dataset.train_dataset[0]
            y = np.array(dataset.train_dataset[1])

        y_pred = np.array(self.model.predict(X))

        self._fit(X, y, y_pred, **kwargs)

    @abstractmethod
    def _test(self, X, y_pred, **kwargs):
        """To be implemented by the child class; tests the method on the provided preprocessed dataset.

        Note: _test shouldn"t be directly called by the user, rather it should be called indirectly via the test method.

        Parameters:
            X (np.ndarray): features, SMILES
            y_pred (np.ndarray): predicted values
        """
        pass

    def test(
        self,
        dataset: Union[BaseDataset, list, str],
        use_entire_dataset: bool = False,
        **kwargs
    ):
        """Evaluates model uncertainty on a test dataset. Calls the _test method on the provided dataset.

        Parameters:
            dataset (BaseDataset, list, str): dataset being tested, sequence of SMILES
            use_entire_dataset (bool): whether to use the entire dataset or only the testing data
        """
        if isinstance(dataset, BaseDataset):
            if use_entire_dataset:
                X = dataset.entire_dataset[0]
            else:
                X = dataset.test_dataset[0]
        elif isinstance(dataset, list):
            X = dataset
        else:
            X = [dataset]

        y_pred = np.array(self.model.predict(X))

        return self._test(X, y_pred, **kwargs)

    def preprocess(self, X, y=None):
        """Preprocesses data into the appropriate representation.

        The preprocess for the model must return a np.ndarray, e.g. the model must use a
        BaseStructVecRepresentation or a representation must be passed

        Parameters:
            X (np.ndarray): features, SMILES
            y (np.ndarray): values

        Returns:
            X (np.ndarray): features, representation
        """
        if self.rep is None:
            X = self.model.preprocess(X, y)
        else:
            X = np.array(self.rep.convert(X))

        assert isinstance(
            X, np.ndarray
        ), "The preprocess for the model must return a np.ndarray, e.g. the model must use a BaseCompoundVecRepresentation or a representation must be passed"

        return X

    def _save(self) -> dict:
        return super()._save()

    def _load(self, d: dict):
        return super()._load(d)



import numpy as np
import torch
import torch_geometric.data
from ogb.utils.features import atom_to_feature_vector, bond_to_feature_vector
from rdkit import Chem

from olorenchemengine.internal import mock_imports


def molecule_to_graph(mol, include_mol=False):
    # Convert to RDKit molecule
    if type(mol) == str:
        mol = Chem.MolFromSmiles(mol)

    # Generate nodes of the graph
    atom_features_list = []
    for atom in mol.GetAtoms():
        atom_features_list.append(atom_to_feature_vector(atom))
    x = np.array(atom_features_list, dtype=np.int64)

    # Generate edges of the graph
    num_bond_features = 3  # bond type, bond stereo, is_conjugated
    if len(mol.GetBonds()) > 0:  # mol has bonds
        edges_list = []
        edge_features_list = []
        for bond in mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()

            edge_feature = bond_to_feature_vector(bond)

            # add edges in both directions
            edges_list.append((i, j))
            edge_features_list.append(edge_feature)
            edges_list.append((j, i))
            edge_features_list.append(edge_feature)

        # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
        edge_index = np.array(edges_list, dtype=np.int64).T

        # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
        edge_attr = np.array(edge_features_list, dtype=np.int64)

    else:  # mol has no bonds
        edge_index = np.empty((2, 0), dtype=np.int64)
        edge_attr = np.empty((0, num_bond_features), dtype=np.int64)

    graph = dict()
    graph["edge_index"] = edge_index
    graph["edge_feat"] = edge_attr
    graph["node_feat"] = x
    graph["num_nodes"] = len(x)

    if include_mol:
        graph["mol"] = mol

    return graph


def molecule_to_pyg(smiles_str, y=None):
    graph = molecule_to_graph(smiles_str)  # construct ogb graph
    g = torch_geometric.data.Data()
    g.__num_nodes__ = graph["num_nodes"]
    g.edge_index = torch.from_numpy(graph["edge_index"])

    del graph["num_nodes"]
    del graph["edge_index"]

    if graph["edge_feat"] is not None:
        g.edge_attr = torch.from_numpy(graph["edge_feat"])
        del graph["edge_feat"]

    if graph["node_feat"] is not None:
        g.x = torch.from_numpy(graph["node_feat"])
        del graph["node_feat"]

    if y is not None:
        if type(y) == bool:
            g.y = torch.LongTensor([1 if y else 0])
        else:
            g.y = torch.FloatTensor([y])

    return g


def batch_molecule_to_pyg(smiles_list, y=None):
    graphs = [molecule_to_pyg(smiles, y) for smiles in smiles_list]
    return torch_geometric.data.Batch.from_data_list(graphs)



""" A library of various molecular representations.
"""

from abc import abstractmethod, abstractproperty
from typing import Any, List, Union

import numpy as np
from pandas.api.types import is_numeric_dtype
from rdkit import Chem
from tqdm import tqdm

import olorenchemengine as oce

from .base_class import *
from .dataset import *


def get_all_reps():
    return list(oce.all_subclasses(BaseRepresentation))


class BaseRepresentation(BaseClass):

    """BaseClass for all molecular representations (PyTorch Geometric graphs, descriptors, fingerprints, etc.)

    Parameters:
        log (boolean): Whether to log the representation or not

    Methods:
        _convert(smiles: str, y: Union[int, float, np.number] = None) -> Any: converts a single structure (represented by a SMILES string) to a representation
        convert(Xs: Union[list, pd.DataFrame, dict, str], ys: Union[list, pd.Series, np.ndarray]=None) -> List[Any]: converts input data to a list of representations
    """

    @log_arguments
    def __init__(self, log=True):
        pass

    @abstractmethod
    def _convert(self, smiles: str, y: Union[int, float, np.number] = None) -> Any:
        """Converts a single structure (represented by a SMILES string) to a representation

        Parameters:
            smiles (str): SMILES string of the structure
            y (Union[int, float, np.number]): target value of the structure

        Returns:
            Any: representation of the structure
        """
        pass

    def _convert_list(
        self, smiles_list: List[str], ys: List[Union[int, float, np.number]] = None
    ) -> List[Any]:
        """Converts a list of structures (represented by a SMILES string) to a list of representations

        Parameters:
            smiles_list (List[str]): list of SMILES strings of the structures
            ys (List[Union[int, float, np.number]]): list of target values of the structures

        Returns:
            List[Any]: list of representations of the structures"""
        if ys is None:
            return [self._convert(s) for s in smiles_list]
            # with mp.Pool(mp.cpu_count()) as p:
            #   X = p.map(self._convert, smiles_list)
            # return X
        else:
            return [self._convert(s, y=y) for s, y in tqdm(zip(smiles_list, ys))]

    def _convert_cache(
        self, smiles: str, y: Union[int, float, np.number] = None
    ) -> Any:
        """Converts a single structure (represented by a SMILES string) to a representation

        Parameters:
            smiles (str): SMILES string of the structure
            y (Union[int, float, np.number]): target value of the structure
        Returns:
            Any: representation of the structure
        """
        if smiles in self.cache.keys():
            return self.cache[smiles]
        else:
            return self._convert(smiles, y=y)

    def convert(
        self,
        Xs: Union[list, pd.DataFrame, dict, str],
        ys: Union[list, pd.Series, np.ndarray] = None,
        **kwargs,
    ) -> List[Any]:
        """Converts input data to a list of representations

        Parameters:
            Xs (Union[list, pd.DataFrame, dict, str]): input data
            ys (Union[list, pd.Series, np.ndarray]=None): target values of the input data

        Returns:
            List[Any]: list of representations of the input data
        """
        if isinstance(Xs, list) and (
            isinstance(Xs[0], list) or isinstance(Xs[0], tuple)
        ):
            smiles = [X[0] for X in Xs]
        elif isinstance(Xs, pd.DataFrame) or isinstance(Xs, dict):
            if isinstance(Xs, pd.DataFrame):
                keys = Xs.columns.tolist()
            elif isinstance(Xs, dict):
                keys = Xs.keys()
            if "smiles" in keys:
                smiles = Xs["smiles"]
            elif "Smiles" in keys:
                smiles = Xs["Smiles"]
            elif "SMILES" in keys:
                smiles = Xs["SMILES"]
            else:
                smiles = Xs.iloc[:, 0].tolist()
        elif isinstance(Xs, str):
            smiles = [Xs]
        else:
            smiles = Xs

        return self._convert_list(smiles, ys=ys)

    def _save(self):
        return {}

    def _load(self, d):
        pass


class SMILESRepresentation(BaseRepresentation):
    """Extracts the SMILES strings from inputted data

    Methods:
        convert(Xs: Union[list, pd.DataFrame, dict, str], ys: Union[list, pd.Series, np.ndarray]=None) -> List[Any]: converts input data to a list of SMILES strings
            Data types:
                pd.DataFrames will have columns "smiles" or "Smiles" or "SMILES" extracted
                lists and tuples of multiple elements will have their first element extracted
                strings will be converted to a list of one element
                everything else will be returned as inputted
    """

    def _convert(self, smiles, y=None):
        pass

    def convert(self, Xs, ys=None, **kwargs):
        if isinstance(Xs, list) and (
            isinstance(Xs[0], list) or isinstance(Xs[0], tuple)
        ):
            smiles = [X[0] for X in Xs]
        elif isinstance(Xs, pd.DataFrame) or isinstance(Xs, dict):
            if isinstance(Xs, pd.DataFrame):
                keys = Xs.columns.tolist()
            elif isinstance(Xs, dict):
                keys = Xs.keys()
            if "smiles" in keys:
                smiles = Xs["smiles"]
            elif "Smiles" in keys:
                smiles = Xs["Smiles"]
            elif "SMILES" in keys:
                smiles = Xs["SMILES"]
            elif "smi" in keys:
                smiles = Xs["smi"]
            else:
                smiles = Xs.iloc[:, 0]
        elif isinstance(Xs, str):
            smiles = [Xs]
        elif isinstance(Xs, pd.Series):
            smiles = Xs.tolist()
        else:
            smiles = Xs

        if isinstance(smiles, pd.Series):
            smiles = smiles.tolist()
        return smiles


class AtomFeaturizer(BaseClass):
    """Abstract class for atom featurizers, which create a vector representation for a single atom.

    Methods:
        length(self) -> int: returns the length of the atom vector representation, to be implemented by subclasses
        convert(self, atom: Chem.Atom) -> np.ndarray: converts a single Chem.Atom string to a vector representation, to be implemented by subclasses
    """

    @abstractproperty
    def length(self) -> int:
        pass

    @abstractmethod
    def convert(self, atom: Chem.Atom) -> np.ndarray:
        pass

    def _save(self) -> dict:
        return {}

    def _load(self, d: dict):
        pass


class BondFeaturizer(BaseClass):
    """Abstract class for bond featurizers, which create a vector representation for a single bond.

    Methods:
        length(self) -> int: returns the length of the bond vector representation, to be implemented by subclasses
        convert(self, bond: Chem.Bond) -> np.ndarray: converts a single Chem.Bond string to a vector representation, to be implemented by subclasses
    """

    @abstractproperty
    def length(self) -> int:
        pass

    @abstractmethod
    def convert(self, bond: Chem.Bond) -> np.ndarray:
        pass

    def _save(self):
        return {}

    def _load(self, d: dict):
        pass


class ConcatenatedAtomFeaturizers(AtomFeaturizer):
    """Concatenates multiple atom featurizers into a single vector representation.

    Methods:
        length(self) -> int: returns the length of the atom vector representation, to be implemented by subclasses
        convert(self, atom: Chem.Atom) -> np.ndarray: converts a single Chem.Atom string to a vector representation, to be implemented by subclasses
    """

    def __init__(self, atom_featurizers: List[AtomFeaturizer]):
        self.atom_featurizers = atom_featurizers

    @property
    def length(self) -> int:
        return sum([f.length for f in self.atom_featurizers])

    def convert(self, atom: Chem.Atom) -> np.ndarray:
        return np.concatenate([f.convert(atom) for f in self.atom_featurizers])

    def _save(self):
        return {"atom_featurizers": [f._save() for f in self.atom_featurizers]}

    def _load(self, d: dict):
        for d_, af in zip(d["atom_featurizers"], self.atom_featurizers):
            af._load(d_)


class ConcatenatedBondFeaturizers(BondFeaturizer):
    """Concatenates multiple bond featurizers into a single vector representation.

    Methods:
        length(self) -> int: returns the length of the bond vector representation, to be implemented by subclasses
        convert(self, bond: Chem.Bond) -> np.ndarray: converts a single Chem.Bond string to a vector representation, to be implemented by subclasses
    """

    def __init__(self, bond_featurizers: List[BondFeaturizer]):
        self.bond_featurizers = bond_featurizers

    @property
    def length(self) -> int:
        return sum([f.length for f in self.bond_featurizers])

    def convert(self, bond: Chem.Bond) -> np.ndarray:
        return np.concatenate([f.convert(bond) for f in self.bond_featurizers])

    def _save(self):
        return {"bond_featurizers": [f._save() for f in self.bond_featurizers]}

    def _load(self, d: dict):
        for d_, bf in zip(d["bond_featurizers"], self.bond_featurizers):
            bf._load(d_)


class OGBAtomFeaturizer(AtomFeaturizer):
    """Creates a vector representation for a single atom using the Open Graph Benchmark's atom_to_feature_vector function."""

    @log_arguments
    def __init__(self):
        pass

    @property
    def length(self):
        return 9

    def convert(self, atom: Chem.Atom):
        from ogb.utils.features import atom_to_feature_vector

        return atom_to_feature_vector(atom)


class OGBBondFeaturizer(BondFeaturizer):
    """Creates a vector representation for a single bond using the Open Graph Benchmark's bond_to_feature_vector function."""

    @log_arguments
    def __init__(self):
        pass

    @property
    def length(self):
        return 3

    def convert(self, bond: Chem.Bond):
        from ogb.utils.features import bond_to_feature_vector

        return bond_to_feature_vector(bond)


class TorchGeometricGraph(BaseRepresentation):
    """Representation which returns torch_geometric.data.Data objects.

    Parameters:
        atom_featurizer (AtomFeaturizer): featurizer for atoms
        bond_featurizer (BondFeaturizer): featurizer for bonds

    Attributes:
        dimensions (Tuple[int, int]): number of dimensions for the atom and bond representations

    Methods:
        _convert(self, smiles: str, y: Any=None) -> Data: converts a single SMILES string to a torch_geometric.data.Data object
    """

    @log_arguments
    def __init__(
        self,
        atom_featurizer: AtomFeaturizer = OGBAtomFeaturizer(),
        bond_featurizer: BondFeaturizer = OGBBondFeaturizer(),
        **kwargs,
    ):

        self.atom_featurizer = atom_featurizer
        self.bond_featurizer = bond_featurizer

    @property
    def dimensions(self):
        return (self.atom_featurizer.length, self.bond_featurizer.length)

    def _convert(self, smiles, y=None, addHs=False, **kwargs):
        from torch import Tensor, from_numpy
        from torch_geometric.data import Data

        data = Data()

        mol = Chem.MolFromSmiles(smiles)
        if addHs:
            mol = Chem.AddHs(mol)

        # atoms
        atom_features_list = []
        for atom in mol.GetAtoms():
            atom_features_list.append(self.atom_featurizer.convert(atom))
        x = np.array(atom_features_list, dtype=np.int64)

        # bonds
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()

                edge_feature = self.bond_featurizer.convert(bond)

                # add edges in both directions
                edges_list.append((i, j))
                edge_features_list.append(edge_feature)
                edges_list.append((j, i))
                edge_features_list.append(edge_feature)

            # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
            edge_index = np.array(edges_list, dtype=np.int64).T

            # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
            edge_attr = np.array(edge_features_list, dtype=np.int64)
        else:  # mol has no bonds
            edge_index = np.empty((2, 0), dtype=np.int64)
            edge_attr = np.empty((0, self.bond_featurizer.length), dtype=np.int64)

        graph = dict()
        graph["edge_index"] = edge_index
        graph["edge_feat"] = edge_attr
        graph["node_feat"] = x
        graph["num_nodes"] = len(x)

        data.__num_nodes__ = int(graph["num_nodes"])
        data.edge_index = from_numpy(graph["edge_index"])
        data.edge_attr = from_numpy(graph["edge_feat"])
        data.x = from_numpy(graph["node_feat"])
        if not y is None:
            data.y = Tensor([y])

        del graph["num_nodes"]
        del graph["edge_index"]
        del graph["edge_feat"]
        del graph["node_feat"]

        return data

    def convert(
        self,
        Xs: Union[list, pd.DataFrame, dict, str],
        ys: Union[list, pd.Series, np.ndarray] = None,
        **kwargs,
    ) -> List[Any]:
        Xs = SMILESRepresentation().convert(Xs)
        return super().convert(Xs, ys=ys, **kwargs)

    def _save(self):
        return {
            "atom_featurizer": self.atom_featurizer._save(),
            "bond_featurizer": self.bond_featurizer._save(),
        }

    def _load(self, d):
        self.atom_featurizer._load(d["atom_featurizer"])
        self.bond_featurizer._load(d["bond_featurizer"])


class BaseVecRepresentation(BaseRepresentation):
    """Representation where given input data, returns a vector representation for each compound."""

    @log_arguments
    def __init__(
        self,
        *args,
        collinear_thresh=1.01,
        scale=StandardScaler(),
        names=None,
        log=True,
        **kwargs,
    ):
        self.collinear_thresh = collinear_thresh
        self.to_drop = None
        if not scale is None:
            scale = scale.copy()
        self.scale = scale
        self.scale_fitted = False
        self._names = names

        from os import path

        if not path.exists(path.join(path.expanduser("~"), f".oce/cache/")):
            os.mkdir(path.join(path.expanduser("~"), f".oce/cache/"))
        if not path.exists(path.join(path.expanduser("~"), f".oce/cache/vecrep/")):
            os.mkdir(path.join(path.expanduser("~"), f".oce/cache/vecrep/"))
        if not path.exists(
            path.join(
                path.expanduser("~"), f".oce/cache/vecrep/{self.__class__.__name__}"
            )
        ):
            os.mkdir(
                path.join(
                    path.expanduser("~"), f".oce/cache/vecrep/{self.__class__.__name__}"
                )
            )

        super().__init__(*args, log=False, **kwargs)

    @property
    def names(self):
        if not self._names is None:
            return self._names
        else:
            raise ValueError(
                f"Names not set for representation {self.__class__.__name__}"
            )

    def convert(
        self,
        Xs: Union[list, pd.DataFrame, dict, str],
        ys: Union[list, pd.Series, np.ndarray] = None,
        fit=False,
        **kwargs,
    ) -> List[np.ndarray]:
        """BaseVecRepresentation's convert returns a list of numpy arrays.

        Args:
            Xs (Union[list, pd.DataFrame, dict, str]): input data
            ys (Union[list, pd.Series, np.ndarray], optional): included for compatibility, unused argument. Defaults to None.

        Returns:
            List[np.ndarray]: list of molecular vector representations
        """
        import joblib

        input_hash = (
            joblib.hash(Xs)
            + joblib.hash(ys)
            + joblib.hash(self._save())
            + joblib.hash(oce.parameterize(self))
        )

        from os import path

        if path.exists(
            path.join(
                path.expanduser("~"),
                f".oce/cache/vecrep/{self.__class__.__name__}/{input_hash}.npy",
            )
        ):
            return np.load(
                path.join(
                    path.expanduser("~"),
                    f".oce/cache/vecrep/{self.__class__.__name__}/{input_hash}.npy",
                ),
                allow_pickle=True,
            )

        feats = super().convert(Xs, ys)
        import pandas as pd

        feats = pd.DataFrame.from_records(
            feats, columns=[f"col{i}" for i in range(len(feats[0]))]
        )
        if fit and len(Xs) > 2:
            # collinear
            corr_matrix = feats.corr().abs()
            upper = corr_matrix.where(
                np.triu(np.ones(corr_matrix.shape), k=1).astype(np.bool)
            )
            self.to_drop = [
                column
                for column in upper.columns
                if any(upper[column] > self.collinear_thresh)
            ]
            feats = feats.drop(columns=self.to_drop)
            # scale
            if not self.scale is None:
                x = self.scale.fit_transform(feats.values)
                feats.values[:] = np.nan_to_num(x.reshape(feats.values.shape))
                self.scale_fitted = True
        else:
            if not self.to_drop is None:
                feats = feats.drop(columns=self.to_drop)
            # scale
            if self.scale_fitted and self.scale is not None:
                x = self.scale.transform(feats.values)
                feats.values[:] = np.nan_to_num(x.reshape(feats.values.shape))

        output = np.nan_to_num(np.array(feats.to_records(index=False).tolist()))
        np.save(
            path.join(
                path.expanduser("~"),
                f".oce/cache/vecrep/{self.__class__.__name__}/{input_hash}.npy",
            ),
            output,
            allow_pickle=True,
        )
        return output

    def calculate_distance(
        self,
        x1: Union[str, List[str]],
        x2: Union[str, List[str]],
        metric: str = "cosine",
        **kwargs,
    ) -> np.ndarray:
        """Calculates the distance between two molecules or list of molecules.

        Returns a 2D array of distances between each pair of molecules of shape
        len(x1) by len(x2).

        This uses pairwise_distances from sklearn.metrics to calculate distances
        between the vector representations of the molecules. Options for distances
        are Valid values for metric are:

            From scikit-learn: [‘cityblock’, ‘cosine’, ‘euclidean’, ‘l1’, ‘l2’,
                ‘manhattan’]. These metrics support sparse matrix inputs.
                [‘nan_euclidean’] but it does not yet support sparse matrices.
            From scipy.spatial.distance: [‘braycurtis’, ‘canberra’, ‘chebyshev’,
                ‘correlation’, ‘dice’, ‘hamming’, ‘jaccard’, ‘kulsinski’,
                ‘mahalanobis’, ‘minkowski’, ‘rogerstanimoto’, ‘russellrao’,
                ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’,
                ‘yule’].

            See the documentation for scipy.spatial.distance for details on these metrics.
        """

        from sklearn.metrics import pairwise_distances

        if isinstance(x1, str):
            x1 = [x1]
        if isinstance(x2, str):
            x2 = [x2]
        x1 = self.convert(x1)
        x2 = self.convert(x2)
        return pairwise_distances(x1, x2, metric=metric, **kwargs)

    def __add__(self, other):
        """Adds two representations together

        Parameters:
            other (BaseVecRepresentation): representation to add to the current representation

        Returns:
            BaseVecRepresentation: new representation"""
        return ConcatenatedVecRepresentation(self, other)

    def _save(self):
        d = {}
        d.update({"collinear_thresh": self.collinear_thresh})
        d.update({"to_drop": self.to_drop})
        if not self.scale is None:
            d.update({"scale": self.scale._save()})
            d.update({"scale_fitted": self.scale_fitted})
        return d

    def _load(self, d):
        self.collinear_thresh = d["collinear_thresh"]
        self.to_drop = d["to_drop"]
        if not self.to_drop is None:
            self.to_drop = set(self.to_drop)
        if not self.scale is None and "scale" in d:
            self.scale._load(d["scale"])
            self.scale_fitted = d["scale_fitted"]


class ConcatenatedVecRepresentation(BaseVecRepresentation):
    """Creates a structure vector representation by concatenating multiple representations.

    Parameters:
        rep1 (BaseVecRepresentation): first representation to concatenate
        rep2 (BaseVecRepresentation): second representation to concatenate
        log (bool): whether to log the representations or not

    Can be created by adding two representations together using + operator.

    Example
    ------------------------------
    import olorenautoml as oam
    combo_rep = oam.MorganVecRepresentation(radius=2, nbits=2048) + oam.Mol2Vec()
    model = oam.RandomForestModel(representation = combo_rep, n_estimators = 1000)

    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        rep1: BaseVecRepresentation,
        rep2: BaseVecRepresentation,
        log=True,
        **kwargs,
    ):
        self.rep1 = rep1
        self.rep2 = rep2
        try:
            self._names = rep1.names + rep2.names
        except Exception as e:
            self._names = None
        super().__init__(names=self._names, log=False, **kwargs)

    def _convert(self, smiles, y=None, fit=False):
        converted_1 = self.rep1._convert(smiles, y=y, fit=fit)
        converted_2 = self.rep2._convert(smiles, y=y, fit=fit)
        return np.concatenate((converted_1, converted_2))

    def _convert_list(self, smiles_list, ys=None, fit=False):
        converted_1 = self.rep1._convert_list(smiles_list, ys=ys, fit=fit)
        converted_2 = self.rep2._convert_list(smiles_list, ys=ys, fit=fit)
        return np.concatenate((converted_1, converted_2), axis=1)

    def convert(self, smiles_list, ys=None, fit=False, **kwargs):
        converted_1 = self.rep1.convert(smiles_list, ys=ys, fit=fit)
        converted_2 = self.rep2.convert(smiles_list, ys=ys, fit=fit)
        return np.concatenate((converted_1, converted_2), axis=1)


class NoisyVec(BaseVecRepresentation):
    """Adds noise to a given BaseVecRepresentation

    Parameters:
        rep (BaseVecRepresentation): BaseVecRepresentation to add noise to
        a_std (float): standard deviation of the additive noise. Defaults to 0.1.
        m_std (float): standard deviation of the multiplicative noise. Defaults to 0.1.
        names (List[str]): list of the names of the features in the vector representation, optional.

    Example
    ------------------------------
    import olorenautoml as oam
    model = oam.RandomForestModel(representation = oam.'''BaseCompoundVecRepresentation(Params)''', n_estimators=1000)

    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self, rep: BaseVecRepresentation, *args, a_std=0.1, m_std=0.1, **kwargs
    ):
        self.a_std = a_std
        self.m_std = m_std
        self.rep = rep
        super().__init__(*args, **kwargs)

    def _convert(self, smiles: str, y=None) -> np.ndarray:
        """Returns specified vector representation of inputted compound with added noise.

        Args:
            smiles (str): SMILES representation of compound
            y (optional): unused argument for compatibility. Defaults to None.

        Returns:
            np.ndarrary: vector representation of inputted compound with added noise
        """
        x = self.rep._convert(smiles)
        x = (x + np.random.normal(0, self.a_std, x.shape)) * np.random.normal(
            1, self.m_std, x.shape
        )
        return x


class DatasetFeatures(BaseVecRepresentation):
    """Selects features from the input dataset as the vector representation"""

    def _convert(self, smiles, y=None):
        pass

    def convert(self, X, **kwargs):
        assert isinstance(X, pd.DataFrame)
        self.numeric_cols = [c for c in X.columns.tolist() if is_numeric_dtype(X[c])]
        assert (
            not len(self.numeric_cols) > 0
        ), "No numeric feature columns found in dataset"
        X = X[self.numeric_cols].to_numpy()
        return X


class BaseCompoundVecRepresentation(BaseVecRepresentation):
    """Computes a vector representation from each structure.

    Parameters:
        normalize (bool): whether to normalize the vector representation or not
        names (List[str]): list of the names of the features in the vector representation, optional."""

    @log_arguments
    def __init__(self, normalize=False, **kwargs):
        self.normalize = normalize
        super().__init__(**kwargs)

    def __len__(self):
        return len(self._convert("C"))

    @abstractmethod
    def _convert(self, smiles, y=None) -> np.ndarray:
        pass

    def convert(
        self,
        Xs: Union[list, pd.Series, pd.DataFrame, dict, str],
        ys: Union[list, pd.Series, np.ndarray] = None,
        fit=False,
        **kwargs,
    ) -> np.ndarray:
        """Computes a vector representation from each structure in Xs."""
        feats = np.array(super().convert(Xs, ys, fit=fit))
        out = np.nan_to_num(feats.astype(np.float32))
        return out

    def inverse(self, Xs):
        """Inverts the vector representation to the original feature values

        Parameters:
            Xs (np.ndarray): vector representation of the structures

        Returns:
            list: list of the original feature values"""

        pass


class ConcatenatedStructVecRepresentation(BaseCompoundVecRepresentation):
    """Creates a structure vector representation by concatenating multiple representations.

    DEPRECEATED, use ConcatenatedVecRepresentation instead.

    Parameters:
        rep1 (BaseVecRepresentation): first representation to concatenate
        rep2 (BaseVecRepresentation): second representation to concatenate
        log (bool): whether to log the representations or not
    """

    @log_arguments
    def __init__(
        self,
        rep1: BaseCompoundVecRepresentation,
        rep2: BaseCompoundVecRepresentation,
        log=True,
        **kwargs,
    ):
        self.rep1 = rep1
        self.rep2 = rep2
        try:
            self._names = rep1.names + rep2.names
        except Exception as e:
            self._names = None
        super().__init__(names=self._names, log=False, **kwargs)

    def _convert(self, smiles, y=None):
        converted_1 = self.rep1._convert(smiles, y=y)
        converted_2 = self.rep2._convert(smiles, y=y)
        return np.concatenate((converted_1, converted_2))

    def _convert_list(self, smiles_list, ys=None):
        converted_1 = self.rep1._convert_list(smiles_list, ys=ys)
        converted_2 = self.rep2._convert_list(smiles_list, ys=ys)
        return np.concatenate((converted_1, converted_2), axis=1)


class DescriptastorusDescriptor(BaseCompoundVecRepresentation):
    """Wrapper for DescriptaStorus descriptors (https://github.com/bp-kelley/descriptastorus)

    Parameters:
        name (str): name of the descriptor. Either "atompaircounts", "morgan3counts",
            "morganchiral3counts","morganfeature3counts","rdkit2d","rdkit2dnormalized",
            "rdkitfpbits"
        log (bool): whether to log the representations or not"""

    available_descriptors = [
        "atompaircounts",
        "morgan3counts",
        "morganchiral3counts",
        "morganfeature3counts",
        "rdkit2d",
        "rdkit2dnormalized",
        "rdkitfpbits",
    ]

    @log_arguments
    def __init__(self, name, *args, log=True, scale=None, **kwargs):
        self.name = name

        from descriptastorus.descriptors.DescriptorGenerator import MakeGenerator

        self.generator = MakeGenerator((name,))
        super().__init__(log=False, scale=scale, **kwargs)

    def _convert(self, smiles, y=None):
        results = self.generator.process(smiles)
        processed, features = results[0], np.nan_to_num(results[1:], nan=0)

        if processed is None:
            print("ERROR: %s" % smiles)
        return features

    @classmethod
    def AllInstances(cls):
        return [cls(name) for name in cls.available_descriptors]


from rdkit.Chem import Lipinski


class LipinskiDescriptor(BaseCompoundVecRepresentation):

    """Wrapper for Lipinski descriptors (https://www.rdkit.org/docs/RDKit_Book.html#Lipinski_Descriptors)

    Parameters:
        log (bool): whether to log the representations or not"""

    @log_arguments
    def __init__(self, log=True, **kwargs):
        super().__init__(
            names=[
                "FractionCSP3",
                "HeavyAtomCount",
                "NHOHCount",
                "NOCount",
                "NumAliphaticCarbocycles",
                "NumAliphaticHeterocycles",
                "NumAliphaticRings",
                "NumAromaticCarbocycles",
                "NumAromaticHeterocycles",
                "NumAromaticRings",
                "NumHAcceptors",
                "NumHDonors",
                "NumHeteroatoms",
                "NumRotatableBonds",
                "NumSaturatedCarbocycles",
                "NumSaturatedHeterocycles",
                "NumSaturatedRings",
                "RingCount",
            ],
            log=False,
            **kwargs,
        )

    def _convert(self, smiles, y=None):
        m = Chem.MolFromSmiles(smiles)
        return [
            Lipinski.FractionCSP3(m),
            Lipinski.HeavyAtomCount(m),
            Lipinski.NHOHCount(m),
            Lipinski.NOCount(m),
            Lipinski.NumAliphaticCarbocycles(m),
            Lipinski.NumAliphaticHeterocycles(m),
            Lipinski.NumAliphaticRings(m),
            Lipinski.NumAromaticCarbocycles(m),
            Lipinski.NumAromaticHeterocycles(m),
            Lipinski.NumAromaticRings(m),
            Lipinski.NumHAcceptors(m),
            Lipinski.NumHDonors(m),
            Lipinski.NumHeteroatoms(m),
            Lipinski.NumRotatableBonds(m),
            Lipinski.NumSaturatedCarbocycles(m),
            Lipinski.NumSaturatedHeterocycles(m),
            Lipinski.NumSaturatedRings(m),
            Lipinski.RingCount(m),
        ]


class FragmentIndicator(BaseCompoundVecRepresentation):
    """Indicator variables for all fragments in rdkit.Chem.Fragments

    http://rdkit.org/docs/source/rdkit.Chem.Fragments.html
    """

    @log_arguments
    def __init__(self, log=True, **kwargs):
        super().__init__(
            names=[
                "Number of aliphatic carboxylic acids ",
                "Number of aliphatic hydroxyl groups  ",
                "Number of aliphatic hydroxyl groups excluding tert-OH",
                "Number of N functional groups attached to aromatics",
                "Number of Aromatic carboxylic acide  ",
                "Number of aromatic nitrogens         ",
                "Number of aromatic amines            ",
                "Number of aromatic hydroxyl groups   ",
                "Number of carboxylic acids           ",
                "Number of carboxylic acids           ",
                "Number of carbonyl O                 ",
                "Number of carbonyl O, excluding COOH ",
                "Number of thiocarbonyl               ",
                "Number of C(OH)CCN-Ctert-alkyl or C(OH)CCNcyclic",
                "Number of Imines                     ",
                "Number of Tertiary amines            ",
                "Number of Secondary amines           ",
                "Number of Primary amines             ",
                "Number of hydroxylamine groups       ",
                "Number of XCCNR groups               ",
                "Number of tert-alicyclic amines (no heteroatoms, not quinine-like bridged N)",
                "Number of H-pyrrole nitrogens        ",
                "Number of thiol groups               ",
                "Number of aldehydes                  ",
                "Number of alkyl carbamates (subject to hydrolysis)",
                "Number of alkyl halides              ",
                "Number of allylic oxidation sites excluding steroid dienone",
                "Number of amides                     ",
                "Number of amidine groups             ",
                "Number of anilines                   ",
                "Number of aryl methyl sites for hydroxylation",
                "Number of azide groups               ",
                "Number of azo groups                 ",
                "Number of barbiturate groups         ",
                "Number of benzene rings              ",
                "Number of benzodiazepines with no additional fused rings",
                "Bicyclic                             ",
                "Number of diazo groups               ",
                "Number of dihydropyridines           ",
                "Number of epoxide rings              ",
                "Number of esters                     ",
                "Number of ether oxygens (including phenoxy)",
                "Number of furan rings                ",
                "Number of guanidine groups           ",
                "Number of halogens                   ",
                "Number of hydrazine groups           ",
                "Number of hydrazone groups           ",
                "Number of imidazole rings            ",
                "Number of imide groups               ",
                "Number of isocyanates                ",
                "Number of isothiocyanates            ",
                "Number of ketones                    ",
                "Number of ketones excluding diaryl, a,b-unsat. dienones, heteroatom on Calpha",
                "Number of beta lactams               ",
                "Number of cyclic esters (lactones)   ",
                "Number of methoxy groups -OCH3       ",
                "Number of morpholine rings           ",
                "Number of nitriles                   ",
                "Number of nitro groups               ",
                "Number of nitro benzene ring substituents",
                "Number of non-ortho nitro benzene ring substituents",
                "Number of nitroso groups, excluding NO2",
                "Number of oxazole rings              ",
                "Number of oxime groups               ",
                "Number of para-hydroxylation sites   ",
                "Number of phenols                    ",
                "Number of phenolic OH excluding ortho intramolecular Hbond substituents",
                "Number of phosphoric acid groups     ",
                "Number of phosphoric ester groups    ",
                "Number of piperdine rings            ",
                "Number of piperzine rings            ",
                "Number of primary amides             ",
                "Number of primary sulfonamides       ",
                "Number of pyridine rings             ",
                "Number of quarternary nitrogens      ",
                "Number of thioether                  ",
                "Number of sulfonamides               ",
                "Number of sulfone groups             ",
                "Number of terminal acetylenes        ",
                "Number of tetrazole rings            ",
                "Number of thiazole rings             ",
                "Number of thiocyanates               ",
                "Number of thiophene rings            ",
                "Number of unbranched alkanes of at least 4 members (excludes halogenated alkanes)",
                "Number of urea groups                ",
            ],
            log=False,
            **kwargs,
        )

    def _convert(self, smiles, y=None):
        from rdkit import Chem

        mol = Chem.MolFromSmiles(smiles)

        from rdkit.Chem import Fragments

        return [
            Fragments.fr_Al_COO(mol),
            Fragments.fr_Al_OH(mol),
            Fragments.fr_Al_OH_noTert(mol),
            Fragments.fr_ArN(mol),
            Fragments.fr_Ar_COO(mol),
            Fragments.fr_Ar_N(mol),
            Fragments.fr_Ar_NH(mol),
            Fragments.fr_Ar_OH(mol),
            Fragments.fr_COO(mol),
            Fragments.fr_COO2(mol),
            Fragments.fr_C_O(mol),
            Fragments.fr_C_O_noCOO(mol),
            Fragments.fr_C_S(mol),
            Fragments.fr_HOCCN(mol),
            Fragments.fr_Imine(mol),
            Fragments.fr_NH0(mol),
            Fragments.fr_NH1(mol),
            Fragments.fr_NH2(mol),
            Fragments.fr_N_O(mol),
            Fragments.fr_Ndealkylation1(mol),
            Fragments.fr_Ndealkylation2(mol),
            Fragments.fr_Nhpyrrole(mol),
            Fragments.fr_SH(mol),
            Fragments.fr_aldehyde(mol),
            Fragments.fr_alkyl_carbamate(mol),
            Fragments.fr_alkyl_halide(mol),
            Fragments.fr_allylic_oxid(mol),
            Fragments.fr_amide(mol),
            Fragments.fr_amidine(mol),
            Fragments.fr_aniline(mol),
            Fragments.fr_aryl_methyl(mol),
            Fragments.fr_azide(mol),
            Fragments.fr_azo(mol),
            Fragments.fr_barbitur(mol),
            Fragments.fr_benzene(mol),
            Fragments.fr_benzodiazepine(mol),
            Fragments.fr_bicyclic(mol),
            Fragments.fr_diazo(mol),
            Fragments.fr_dihydropyridine(mol),
            Fragments.fr_epoxide(mol),
            Fragments.fr_ester(mol),
            Fragments.fr_ether(mol),
            Fragments.fr_furan(mol),
            Fragments.fr_guanido(mol),
            Fragments.fr_halogen(mol),
            Fragments.fr_hdrzine(mol),
            Fragments.fr_hdrzone(mol),
            Fragments.fr_imidazole(mol),
            Fragments.fr_imide(mol),
            Fragments.fr_isocyan(mol),
            Fragments.fr_isothiocyan(mol),
            Fragments.fr_ketone(mol),
            Fragments.fr_ketone_Topliss(mol),
            Fragments.fr_lactam(mol),
            Fragments.fr_lactone(mol),
            Fragments.fr_methoxy(mol),
            Fragments.fr_morpholine(mol),
            Fragments.fr_nitrile(mol),
            Fragments.fr_nitro(mol),
            Fragments.fr_nitro_arom(mol),
            Fragments.fr_nitro_arom_nonortho(mol),
            Fragments.fr_nitroso(mol),
            Fragments.fr_oxazole(mol),
            Fragments.fr_oxime(mol),
            Fragments.fr_para_hydroxylation(mol),
            Fragments.fr_phenol(mol),
            Fragments.fr_phenol_noOrthoHbond(mol),
            Fragments.fr_phos_acid(mol),
            Fragments.fr_phos_ester(mol),
            Fragments.fr_piperdine(mol),
            Fragments.fr_piperzine(mol),
            Fragments.fr_priamide(mol),
            Fragments.fr_prisulfonamd(mol),
            Fragments.fr_pyridine(mol),
            Fragments.fr_quatN(mol),
            Fragments.fr_sulfide(mol),
            Fragments.fr_sulfonamd(mol),
            Fragments.fr_sulfone(mol),
            Fragments.fr_term_acetylene(mol),
            Fragments.fr_tetrazole(mol),
            Fragments.fr_thiazole(mol),
            Fragments.fr_thiocyan(mol),
            Fragments.fr_thiophene(mol),
            Fragments.fr_unbrch_alkane(mol),
            Fragments.fr_urea(mol),
        ]


from rdkit.Chem import Descriptors, rdMolDescriptors

from .external import calc_pI


class PeptideDescriptors1(BaseCompoundVecRepresentation):
    @log_arguments
    def __init__(self, log=True, **kwargs):
        self.calc_pI = calc_pI(ph=7.4)
        super().__init__(
            names=["NumAtoms", "MW", *self.calc_pI.names, "NumAmideBonds"],
            log=False,
            **kwargs,
        )

    def _convert(self, smiles, y=None):
        m = Chem.MolFromSmiles(smiles)
        calc_pI_props = self.calc_pI._convert(smiles).tolist()
        x = np.nan_to_num(
            np.array(
                [
                    m.GetNumAtoms(),
                    Descriptors.MolWt(m),
                    *calc_pI_props,
                    rdMolDescriptors.CalcNumAmideBonds(m),
                ]
            ).astype(float)
        )
        return x


from rdkit.Chem import AllChem


class MorganVecRepresentation(BaseCompoundVecRepresentation):
    @log_arguments
    def __init__(self, radius=2, nbits=1024, log=True, **kwargs):
        self.radius = radius
        self.nbits = nbits
        super().__init__(log=False, **kwargs)

    def _convert(self, smiles):
        m = Chem.MolFromSmiles(smiles)
        fp = AllChem.GetMorganFingerprintAsBitVect(m, self.radius, nBits=self.nbits)

        fp_np = np.array(fp)
        return fp_np

    def info(self, smiles):
        info = {}
        m = Chem.MolFromSmiles(smiles)
        fp = AllChem.GetMorganFingerprintAsBitVect(
            m, self.radius, nBits=self.nbits, bitInfo=info
        )
        return info


class MACCSKeys(BaseCompoundVecRepresentation):
    """Calculate MACCS (Molecular ACCess System) Keys fingerprint.

    Durant, Joseph L., et al. "Reoptimization of MDL keys for use in drug discovery."
    Journal of chemical information and computer sciences 42.6 (2002): 1273-1280."""

    @log_arguments
    def __init__(self):
        super().__init__(log=False)

    def _convert(self, s: str) -> np.ndarray:
        return AllChem.GetMACCSKeysFingerprint(Chem.MolFromSmiles(s))


class PubChemFingerprint(BaseCompoundVecRepresentation):
    """PubChem Fingerprint.
    The PubChem fingerprint is a 881 bit structural key,
    which is used by PubChem for similarity searching.
    References
    ----------
    .. [1] ftp://ftp.ncbi.nlm.nih.gov/pubchem/specifications/pubchem_fingerprints.pdf
    .. [2] https://github.com/deepchem/deepchem/blob/master/deepchem/feat/molecule_featurizers/pubchem_fingerprint.py
    Note
    -----
    PubChemPy uses REST API to get the fingerprint, so you need internet access.

    PubChem fingerprint calculations fail for certain compounds (i.e. 'CCCCCCC[NH+](CC)CCCC(O)c1ccc(NS(C)(=O)=O)cc1').
    Certain SMILES strings encountered in datasets are not recognized by PubChem.
    In these cases, the fingerprint is returned with all bits equal to zero.
    """

    @log_arguments
    def __init__(self):
        super().__init__(log=False)

    def _convert(self, s: str) -> np.ndarray:
        oce.import_or_install("pubchempy")

        import pubchempy as pcp

        # Check if retrieval of compound and subsequent descriptor calculation succeed without error
        try:
            pubchem_compound = pcp.get_compounds(s, "smiles")[0]
            fp = [int(bit) for bit in pubchem_compound.cactvs_fingerprint]
        except:
            fp = [0] * 881
        return np.array(fp)


class MordredDescriptor(BaseCompoundVecRepresentation):

    """Wrapper for Mordred descriptors (https://github.com/mordred-descriptor/mordred)

    Parameters:
        log (bool): whether to log the representations or not
        descriptor_set (str): name of the descriptor set to use
        normalize (bool): whether to normalize the descriptors or not
    """

    @log_arguments
    def __init__(
        self,
        descriptor_set: Union[str, list] = "all",
        log: bool = True,
        normalize: bool = False,
        **kwargs,
    ):
        oce.import_or_install("mordred")

        from mordred import Calculator, descriptors

        if descriptor_set == "all":
            self.calc = Calculator(descriptors, ignore_3D=False)

        self.normalize = normalize

        super().__init__(log=False, normalize=self.normalize, **kwargs)

    def _convert(self, smiles, y=None):
        pass

    def convert(self, Xs, ys=None, **kwargs):
        """Convert list of SMILES to descriptors in the form of a numpy array.

        Parameters:
            Xs (list): List of SMILES strings.
            ys (list): List of labels.
            normalize (bool): Whether to normalize the descriptors.

        Returns:
            np.ndarray: Array of descriptors.
            Shape: (len(Xs), len(self.names))"""

        smiles = SMILESRepresentation().convert(Xs, ys=ys)
        mols = [Chem.MolFromSmiles(smi) for smi in smiles]
        feats = pd.DataFrame([self.calc(mol).asdict() for mol in mols])
        for col in feats.columns:
            feats[col] = pd.to_numeric(feats[col], errors="coerce").fillna(0)
        feats = np.nan_to_num(feats.to_numpy())
        return feats


from rdkit import Chem, DataStructs, RDConfig
from rdkit.Chem import AllChem
from rdkit.Chem.Pharm2D import Generate, Gobbi_Pharm2D


class GobbiPharma2D(BaseCompoundVecRepresentation):

    """2D Gobbi pharmacophore descriptor (implemented in RDKit, from https://doi.org/10.1002/(SICI)1097-0290(199824)61:1<47::AID-BIT9>3.0.CO;2-Z)"""

    def _convert(self, smiles, y=None):
        m = Chem.MolFromSmiles(smiles)
        AllChem.EmbedMolecule(m)  # gen 3d

        factory = Gobbi_Pharm2D.factory
        # calc 2d p4 fp
        fp = Generate.Gen2DFingerprint(m, factory)
        # dMat option is Key!
        return fp


class GobbiPharma3D(BaseCompoundVecRepresentation):

    """3D Gobbi pharmacophore descriptor (implemented in RDKit, from https://doi.org/10.1002/(SICI)1097-0290(199824)61:1<47::AID-BIT9>3.0.CO;2-Z)"""

    def _convert(self, smiles, y=None):
        m = Chem.MolFromSmiles(smiles)
        s = AllChem.EmbedMolecule(m, useRandomCoords=True)

        factory = Gobbi_Pharm2D.factory
        if s == -1:
            fp = Generate.Gen2DFingerprint(m, factory)
        else:
            fp = Generate.Gen2DFingerprint(m, factory, dMat=Chem.Get3DDistanceMatrix(m))

        return fp


from collections import OrderedDict

import torch
import torch_geometric.data
from rdkit import Chem
from torch_geometric.data import DataLoader as PyGDataLoader


class OlorenCheckpoint(BaseCompoundVecRepresentation):
    """Use OlorenVec from checkpoint as a molecular representation

    Parameters:
        model_path (str): path to checkpoint file for OlorenVec. Use "default" if unsure
        num_tasks (int): number of coordinates in the vector representation
        log (bool, optional): Log arguments or not. Should only be true if it is not nested. Defaults to True.
    """

    @log_arguments
    def __init__(
        self, model_path: str, num_tasks: int = 2048, log: bool = True, **kwargs
    ):
        if model_path == "default":
            import os
            from os.path import expanduser

            path = os.path.join(expanduser("~"), ".oce/olorenvec.ckpt")

            if not os.path.exists(path):
                print("Downloading OlorenVec checkpoint...")
                import gcsfs

                fs = gcsfs.GCSFileSystem()
                with fs.open("gs://oloren-public-data/saves/olorenvec.ckpt", "rb") as f:
                    with open(path, "wb") as out:
                        out.write(f.read())
        else:
            path = model_path

        state_dict = OrderedDict(
            [
                (k.replace("model.", ""), v)
                for k, v in torch.load(path, map_location=oce.CONFIG["DEVICE"])[
                    "state_dict"
                ].items()
            ]
        )

        from olorenchemengine.pyg.gcn import GNN

        self.model = GNN(
            gnn_type="gcn",
            num_tasks=num_tasks,
            num_layer=5,
            emb_dim=300,
            drop_ratio=0.5,
            virtual_node=False,
        )

        self.model.to(oce.CONFIG["DEVICE"])

        self.model.load_state_dict(state_dict)
        self.model.eval()

        super().__init__(log=False, **kwargs)

    def _convert(self, smiles, y=None):
        try:
            from torch_geometric.data import Batch

            graph = self.smiles2pyg(smiles, y)
            batch = Batch.from_data_list([graph])
            batch.to(oce.CONFIG["DEVICE"])
            x = self.model(batch).detach().cpu().numpy()
            x = np.squeeze(x)
            return x
        except Exception as e:
            print(e)
            print("ERROR ON %s" % smiles)
            return self._convert("C1CCCCC1", y=y)

    def _convert_list(self, smiles_list, ys=None):
        xs = [self.smiles2pyg(s, None) for s in smiles_list]

        kwargs = (
            {"num_workers": oce.CONFIG["NUM_WORKERS"], "pin_memory": True}
            if oce.CONFIG["USE_CUDA"]
            else {}
        )

        dataloader = PyGDataLoader(xs, batch_size=64, **kwargs)

        predictions = list()
        for batch in dataloader:
            batch.to(oce.CONFIG["DEVICE"])
            predictions.append(self.model(batch))

        predictions = torch.cat(predictions, dim=0)
        predictions = predictions.detach().cpu().numpy()
        return predictions

    @classmethod
    def AllInstances(cls):
        return [cls("default", num_tasks=2048)]

    def molecule2graph(self, mol, include_mol=False):
        """Convert a molecule to a PyG graph with features and labels

        Parameters:
            mol (rdkit.Chem.rdchem.Mol): molecule to convert
            include_mol (bool, optional): Whether or not include the molecule in the graph. Defaults to False.

        Returns:
           graph: PyG graph"""

        # Convert to RDKit molecule
        if not isinstance(mol, Chem.Mol):
            mol = Chem.MolFromSmiles(mol)

        from ogb.utils.features import atom_to_feature_vector, bond_to_feature_vector

        # Generate nodes of the graph
        atom_features_list = []
        for atom in mol.GetAtoms():
            atom_features_list.append(atom_to_feature_vector(atom))
        x = np.array(atom_features_list, dtype=np.int64)

        # Generate edges of the graph
        num_bond_features = 3  # bond type, bond stereo, is_conjugated
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()

                edge_feature = bond_to_feature_vector(bond)

                # add edges in both directions
                edges_list.append((i, j))
                edge_features_list.append(edge_feature)
                edges_list.append((j, i))
                edge_features_list.append(edge_feature)

            # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
            edge_index = np.array(edges_list, dtype=np.int64).T

            # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
            edge_attr = np.array(edge_features_list, dtype=np.int64)

        else:  # mol has no bonds
            edge_index = np.empty((2, 0), dtype=np.int64)
            edge_attr = np.empty((0, num_bond_features), dtype=np.int64)

        graph = dict()
        graph["edge_index"] = edge_index
        graph["edge_feat"] = edge_attr
        graph["node_feat"] = x
        graph["num_nodes"] = len(x)

        if include_mol:
            graph["mol"] = mol

        return graph

    def smiles2pyg(self, smiles_str, y, morgan_params={"radius": 2, "nBits": 1024}):
        """Convert a SMILES string to a PyG graph with features and labels

        Parameters:
            smiles_str (str): SMILES string to convert
            y (int): label of the molecule
            morgan_params (dict, optional): parameters for morgan fingerprint. Defaults to {"radius": 2, "nBits": 1024}.

        Returns:
            graph: PyG graph
        """
        graph = self.molecule2graph(smiles_str)  # construct ogb graph
        g = torch_geometric.data.Data()
        g.__num_nodes__ = graph["num_nodes"]
        g.edge_index = torch.from_numpy(graph["edge_index"])

        del graph["num_nodes"]
        del graph["edge_index"]

        if graph["edge_feat"] is not None:
            g.edge_attr = torch.from_numpy(graph["edge_feat"])
            del graph["edge_feat"]

        if graph["node_feat"] is not None:
            g.x = torch.from_numpy(graph["node_feat"])
            del graph["node_feat"]

        if y is None:
            return g

        if type(y) == bool:
            g.y = torch.LongTensor([1 if y else 0])
        else:
            g.y = torch.FloatTensor([y])

        return g


class MCSClusterRep(BaseCompoundVecRepresentation):
    """Clusters a train set of compounds and then finds the maximum common
    substructure (MCS) within each set. The presence of each cluster's MCS is
    used as a feature
    """

    @log_arguments
    def __init__(
        self,
        dataset: BaseDataset,
        *args,
        eval_set="train",
        timeout: int = 5,
        threshold: float = 0.9,
        cached=False,
        log=True,
        **kwargs,
    ):
        if not self.kwargs["cached"]:
            from rdkit import DataStructs
            from rdkit.Chem import AllChem

            if eval_set == "train":
                eval_set = dataset.data[dataset.data["split"] == "train"]
            elif eval_set == "valid":
                eval_set = dataset.data[dataset.data["split"] == "valid"]
            elif eval_set == "test":
                eval_set = dataset.data[dataset.data["split"] == "test"]
            else:
                eval_set = dataset.data

            fingerprinter = lambda s: AllChem.GetMorganFingerprint(
                Chem.MolFromSmiles(s), 4
            )
            eval_set["fps"] = eval_set[dataset.structure_col].apply(fingerprinter)
            corr = np.array(
                [
                    DataStructs.BulkTanimotoSimilarity(fp, eval_set["fps"].tolist())
                    for fp in eval_set["fps"]
                ]
            )

            from rdkit.Chem.rdFMCS import FindMCS
            from sklearn.cluster import AgglomerativeClustering

            clustering = AgglomerativeClustering(
                n_clusters=None,
                distance_threshold=0.4,
                affinity="precomputed",
                linkage="single",
            ).fit(1 - corr)
            eval_set["cluster_id"] = clustering.labels_

            self.smarts = list()
            for cluster_id in tqdm(eval_set["cluster_id"].unique()):
                data_ = eval_set[eval_set["cluster_id"] == cluster_id]
                if len(data_) < 2:
                    continue
                data_["mols"] = data_[dataset.structure_col].apply(Chem.MolFromSmiles)
                mcs = FindMCS(
                    data_["mols"].tolist(), threshold=threshold, timeout=timeout
                )
                self.smarts.append(mcs.smartsString)
            self.args = [""]
            self.kwargs["cached"] = True
        super().__init__(*args, **kwargs, log=False)

    def _convert(self, smiles):
        return np.array(
            [
                Chem.MolFromSmiles(smiles).HasSubstructMatch(Chem.MolFromSmarts(s))
                for s in self.smarts
            ]
        )

    def _save(self):
        d = super()._save()
        d["smarts"] = self.smarts
        return d

    def _load(self, d):
        self.smarts = d["smarts"]
        super()._load(d)


class ModelAsRep(BaseCompoundVecRepresentation):
    """Uses a trained model itself as a representation.

    If we are trying to predict property A, and there is a highly related property B
    that has a lot of data we could train a model on property B and use that model
    with ModelAsRep as a representation for property A.

    Parameters:
        model (BaseModel, str): A trained model to be used as the representation,
            either a BaseModel object or a path to a saved model
        download_public_file (bool, optional): If True, will download the specified
            model from OCE's public warehouse of models. Defaults to False.
        name (str): Name of the property the passed model predicts, which
            is usefully for clear save files/interpretability visualizations.
             Optional.
    """

    @log_arguments
    def __init__(
        self,
        model: Union[BaseModel, str],
        name="ModelAsRep",
        download_public_file=False,
        log=True,
        **kwargs,
    ):
        if isinstance(model, str):
            if download_public_file:
                self.model = oce.load(oce.download_public_file(model))
            else:
                self.model = oce.load(model)
        else:
            self.model = model
        super().__init__(log=False, names=[name], **kwargs)

    def _convert(self, smiles, y=None):
        x = self.model.predict([smiles])
        return x

    def _convert_list(self, smiles_list, ys=None):
        x = np.expand_dims(self.model.predict(smiles_list), axis=1)
        return x

    def _save(self):
        return super()._save()

    def _load(self, d):
        super()._load(d)



""" Ensembling methods to combine `BaseModel`s to create better, combined models.
"""
from .base_class import *
from .basics import *
from .representations import *


def get_oof(self, model, X, y, kf):
    # https://www.kaggle.com/code/arthurtok/introduction-to-ensembling-stacking-in-python/notebook
    oof_train = np.zeros((len(X),))

    for train_index, test_index in kf.split(X):
        x_tr = X.iloc[train_index]
        y_tr = y.iloc[train_index]
        x_te = X.iloc[test_index]

        model.fit(x_tr, y_tr)

        oof_train[test_index] = model.predict(x_te)

    return oof_train


class Averager(BaseModel):
    """Averager averages the predictions of multiple models for an ensembled prediction.

    Parameters:
        models (list): list of BaseModel objects to be averaged.
        n (int, optional): Number of times to repeat the given models. Defaults to 1.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.Averager(models = [
        oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
        oce.SupportVectorMachine(representation = oce.Mol2Vec())
    ])
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(self, models: List[BaseModel], n: int = 1, log: bool = True, **kwargs):
        """BaseStacking constructor

        Args:
            models (List[BaseModel]): List of models to use for the learners to be stacked together.
            stacker_model (BaseModel):a model to use for stacking the models.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor,
                should be True only for unnested classes. Defaults to True.
        """

        if issubclass(type(models), BaseModel):
            models = [models]
        self.models = list()
        for i in range(n):
            for model in models:
                self.models.append(model)
        super().__init__(log=False, **kwargs)

    def preprocess(self, X, y, fit=False):
        """Preprocesses the data for the model.

        Parameters:
            X (pd.DataFrame): Dataframe of features.
            y (pd.DataFrame): Dataframe of labels.

        Returns:
            X (pd.DataFrame): Dataframe of features."""
        return X

    def _fit(self, X, y, valid=None):
        for model in self.models:
            if issubclass(type(model), BaseModel):
                model.fit(X, y)

    def _predict(self, X):
        results = np.zeros(X.shape[0])
        for model in self.models:
            results += model.predict(X) / len(self.models)
        return results

    def _save(self):
        d = super()._save()
        model_save_list = list()
        for i, model in enumerate(self.models):
            model_save_list.append(model._save())
        d.update({"model_save_dict": model_save_list})
        return d

    def _load(self, d):
        super()._load(d)
        for i, model in enumerate(self.models):
            model._load(d["model_save_dict"][i])


class BaseStacking(BaseModel):

    """BaseStacking stacks the predictions of models for an ensembled prediction.

    Parameters:
        models (List[BaseModel]): list of models to use for the learners to be stacked together.
        stacker_model (BaseModel): a model to use for stacking the models.

    Called only by child classes. Not to be called directly by user.
    """

    @log_arguments
    def __init__(
        self,
        models: List[BaseModel],
        stacker_model: BaseModel,
        n: int = 1,
        log: bool = True,
        oof=False,
        nfolds=5,
        **kwargs
    ):
        """BaseStacking constructor

        Args:
            models (List[BaseModel]): List of models to use for the learners to be stacked together.
            stacker_model (BaseModel):a model to use for stacking the models.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor,
                should be True only for unnested classes. Defaults to True.
        """

        if issubclass(type(models), BaseModel):
            models = [models]
        self.models = list()
        for i in range(n):
            for model in models:
                self.models.append(model)
        self.stacker_model = stacker_model
        self.oof = oof
        self.nfolds = nfolds
        super().__init__(log=False, **kwargs)

    def preprocess(self, X, y, fit=False):
        return X

    def featurize(self, X):
        """
        Featurizes the data for the model.

        Parameters:
            X (pd.DataFrame): Dataframe of features.
            y (pd.DataFrame): Dataframe of labels.

        Returns:
            data: featurized dataset."""
        data = []
        for i, model in enumerate(self.models):
            if issubclass(type(model), BaseModel):
                data.append(np.squeeze(np.array(model.predict(X))))
            elif issubclass(type(model), BaseRepresentation):
                data_ = model.convert(X)
                for j in range(len(data_[0])):
                    row = np.squeeze(np.array([d[j] for d in data_]))
                    data.append(row)
        data = np.array(data)
        data = np.transpose(np.squeeze(data))
        if data.ndim == 1:
            data = data.reshape(-1, 1)
        return data

    def _fit(self, X, y, valid=None):
        if not self.oof:
            for model in self.models:
                if issubclass(type(model), BaseModel):
                    model.fit(X, y)
            if valid is None:
                data = self.featurize(X)
                self.stacker_model.fit(data, y)
            else:
                X_valid, y_valid = valid
                data = self.featurize(X_valid)
                self.stacker_model.fit(data, y_valid)
        else:
            from sklearn.model_selection import KFold

            kf = KFold(n_splits=self.nfolds)

            data = []
            for model in self.models:
                if issubclass(type(model), BaseModel):
                    feats = get_oof(model, X, y, kf)
                    data.append(np.squeeze(np.array(feats)))
                elif issubclass(type(model), BaseRepresentation):
                    data_ = model.convert(X)
                    for j in range(len(data_[0])):
                        row = np.squeeze(np.array([d[j] for d in data_]))
                        data.append(row)
            data = np.array(data)
            data = np.transpose(np.squeeze(data))
            if data.ndim == 1:
                data = data.reshape(-1, 1)

            self.stacker_model.fit(data, y)

            for model in self.models:
                if issubclass(type(model), BaseModel):
                    model.fit(X, y)

    def _predict(self, X):
        data = self.featurize(X)
        if hasattr(self.stacker_model, "predict_proba"):
            return self.stacker_model.predict_proba(data)
        else:
            return self.stacker_model.predict(data)

    def _save(self):
        d = super()._save()
        model_save_list = list()
        for i, model in enumerate(self.models):
            model_save_list.append(model._save())
        d.update({"model_save_dict": model_save_list})
        d.update({"stacker_model": self.stacker_model._save()})
        return d

    def _load(self, d):
        super()._load(d)
        for i, model in enumerate(self.models):
            model._load(d["model_save_dict"][i])
        self.stacker_model._load(d["stacker_model"])


class BestStacker(BaseStacking):

    """BestStacker is a stacking method that uses the best model from a collection of models to make an ensembled prediction.

    Parameters:
        models (List[BaseModel]): list of models to use for the learners to be stacked together.
        n (int, optional): Number of times to repeat the given models. Defaults to 1.
        log (bool, optional): Whether or not to log the arguments of this constructor. Defaults to True.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.BestStacker(models = [
        oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
        oce.SupportVectorMachine(representation = oce.Mol2Vec())
    ])
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self, models: List[BaseModel], n: int = 1, k: int = 1, log: bool = True
    ):
        """BestStacker constructor

        Args:
            models (List[BaseModel]): List of models to use for the learners to be stacked together.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            k (int, optional): Number of "best" models to choose for the stacking. Defaults to 1."""

        super().__init__(models, KBestLinearRegression(k=k), n=n, log=False)


class LinearRegressionStacker(BaseStacking):

    """LinearRegressionStacker is a stacking method that uses linear regression on the predictions
        from a collection of models to make an ensembled prediction.

        Parameters:
            models (List[BaseModel]): list of models to use for the learners to be stacked together.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor. Defaults to True.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.LinearRegressionStacker(models = [
        oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
        oce.SupportVectorMachine(representation = oce.Mol2Vec())
    ])
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(self, models: List[BaseModel], n: int = 1, log: bool = True):
        """LinearRegressionStacker constructor

        Args:
            models (List[BaseModel]): List of models to use for the learners to be stacked together.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor,
        """
        super().__init__(models, LinearRegression(), n=n, log=False)


class SKLearnStacker(BaseSKLearnModel):

    """SKLearnStacker is a stacking method that uses a sklearn-like models to make an ensembled prediction.

    Attributes:
        reg_stack: a sklearn-like model to use for stacking the models for regression tasks.
        class_stack: a sklearn-like model to use for stacking the models for classification tasks.

    Called only by child classes. Not to be called directly by user.
    """

    @log_arguments
    def __init__(
        self,
        models: List[BaseModel],
        regression_stacker_model: BaseEstimator,
        classification_stacker_model: BaseEstimator,
        n: int = 1,
        log: bool = True,
    ):
        """SKLearnStacker constructor

        Args:
            models (List[BaseModel]): List of models to use for the learners to be stacked together.
            regression_stacker_model (BaseEstimator): a sklearn-like model to use for stacking the models for regression tasks.
            classification_stacker_model (BaseEstimator): a sklearn-like model to use for stacking the models for classification tasks.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor,
                should be True only for unnested classes. Defaults to True."""
        self.reg_stack = BaseStacking(models, regression_stacker_model, n=n)
        self.class_stack = BaseStacking(models, classification_stacker_model, n=n)
        super().__init__(None, self.reg_stack, self.class_stack, log=False)


class RFStacker(SKLearnStacker):

    """RFStacker is a subclass of SKLearnStacker that uses a random forest models to make an ensembled prediction.

        Parameters:
            models (List[BaseModel]): list of models to use for the learners to be stacked together.
            n_estimators (int, optional): Number of trees in the forest. Defaults to 100.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor. Defaults to True.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.RFStacker(
        models = [
            oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
            oce.SupportVectorMachine(representation = oce.Mol2Vec())],
        n_estimators = 100
    )
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        models: List[BaseModel],
        n_estimators: int = 100,
        n: int = 1,
        log: bool = True,
    ):
        """RFStacker constructor

        Args:
            models (List[BaseModel]): List of models to use for the learners to be stacked together.
            n_estimators (int, optional): Number of trees in the forest. Defaults to 100.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor,
        """
        rf_regressor = RandomForestRegressor(
            n_estimators=n_estimators,
            max_features="log2",
            min_samples_split=4,
            min_samples_leaf=4,
        )
        rf_classifier = RandomForestClassifier(
            max_features="log2",
            min_samples_split=4,
            min_samples_leaf=4,
            n_estimators=n_estimators,
            n_jobs=-1,
            criterion="entropy",
            class_weight="balanced_subsample",
        )

        super().__init__(models, rf_regressor, rf_classifier, n=n, log=False)


class MLPStacker(SKLearnStacker):

    """MLPStacker is a subclass of SKLearnStacker that uses a multi-layer perceptron model to make an ensembled prediction.

        Parameters:
            models (List[BaseModel]): list of models to use for the learners to be stacked together.
            layer_dims (List[int]): list of layer dimensions for the MLP.
            activation (str, optional): activation function to use for the MLP. Defaults to 'tanh'.
            epochs (int, optional): number of epochs to train the MLP. Defaults to 100.
            batch_size (int, optional): batch size for the MLP. Defaults to 16.
            verbose (int, optional): verbosity level for the MLP. Defaults to 0.
            n (int, optional): Number of times to repeat the given models. Defaults to 1.
            log (bool, optional): Whether or not to log the arguments of this constructor. Defaults to True.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.MLPStacker(
        models = [
            oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
            oce.SupportVectorMachine(representation = oce.Mol2Vec())],
        layer_dims = [32, 32],
        activation = 'tanh',
        epochs = 15,
        batch_size = 16,
        verbose = 0,
    )
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        models,
        layer_dims=[2048, 512, 128],
        activation="tanh",
        epochs=100,
        batch_size=16,
        verbose=0,
        n=1,
        log=True,
    ):

        super().__init__(
            models,
            TorchMLP(
                None,
                layer_dims=layer_dims,
                activation=activation,
                epochs=epochs,
                batch_size=batch_size,
                verbose=0,
            ),
            TorchMLP(
                None,
                layer_dims=layer_dims,
                activation=activation,
                epochs=epochs,
                batch_size=batch_size,
                verbose=0,
            ),
            n=n,
            log=False,
        )


class BaseBoosting(BaseModel):
    """BaseBoosting uses models in a gradient boosting fashion to create an ensembled model.

    Parameters:
        models (List[BaseModel]): list of models to use for the learners to be stacked together.
        n (int, optional): Number of times to repeat the given models. Defaults to 1.
        oof (bool, optional): Whether or not to use out-of-fold predictions for the ensembled model. Defaults to False.
        log (bool, optional): Whether or not to log the arguments of this constructor. Defaults to True.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.BaseBoosting(
        models = [
            oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
            oce.SupportVectorMachine(representation = oce.Mol2Vec())]
    )
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        models: List[BaseModel],
        n: int = 1,
        oof=False,
        nfolds=5,
        log: bool = True,
        **kwargs
    ):
        """

        Args:
            models (List[BaseModel]): List of models to use in gradient boosting.
            n (int, optional): Number of times to replicate the inputted models. Defaults to 1.
            oof (bool, optional): Whether to use out of fold predictions for boosting. Defaults to False.
            log (bool, optional): Log arguments or not. Should only be true if it is not nested. Defaults to True.
        """

        if issubclass(type(models), BaseModel):
            models = [models]

        self.models = []

        for i in range(n):
            for model in models:
                self.models.append(model.copy())

        self.oof = oof
        self.nfolds = nfolds

        super().__init__(log=False, **kwargs)

    def _fit(self, X_train, y_train):
        if self.oof:
            from sklearn.model_selection import KFold

            kf = KFold(n_splits=self.nfolds)

        for i, model in enumerate(self.models):
            if self.oof:
                y_pred = get_oof(model, X_train, y_train, kf)
                model.fit(X_train, y_train)
            else:
                model.fit(X_train, y_train)
                y_pred = np.array(model.predict(X_train)).flatten()
            y_train = y_train - y_pred

    def _error_calc(self, residuals):
        initial_residual, other_residuals = residuals[0], residuals[1:]
        y = np.array(([initial_residual]))
        for index in range(len(other_residuals)):
            curr_residual = other_residuals[index]
            total_residual = y[-1] - curr_residual
            y = np.append(y, [total_residual], axis=0)

        stdevs = list(np.std(y, axis=1))
        return stdevs

    def _waterfall(self, y_data=None, normalize=False):
        self.stdevs = self._error_calc(self.residuals)
        if y_data is not None:
            naive_mean = np.average(y_data)
            naive_residuals = [
                (y - naive_mean) / np.std(y_data) for y in y_data
            ]  # normalized so they're in the same format as predict's input
            if normalize:
                naive_residuals = naive_residuals
            else:
                naive_residuals = self._unnormalize(np.array(naive_residuals))
            self.residuals.insert(0, naive_residuals)
            self.stdevs.insert(0, np.std(naive_residuals))
        return self.stdevs

    def _predict(self, X, waterfall=False, normalize=False):
        if waterfall:
            self.residuals = []
        y = np.zeros((len(X)))
        for model in self.models:
            prediction = np.array(model.predict(X)).flatten()
            y = y + prediction
            if waterfall:
                if normalize:
                    self.residuals.append(prediction)
                else:
                    self.residuals.append(self._unnormalize(prediction))
        if self.setting == "classification":
            y = np.clip(y, 0, 1)
            if waterfall:
                self.residuals = [
                    np.clip(residual, 0, 1) for residual in self.residuals
                ]
        return y.flatten()

    def _save(self):
        d = super()._save()
        model_save_list = list()
        for i, model in enumerate(self.models):
            model_save_list.append(model._save())
        d.update({"model_save_dict": model_save_list})
        return d

    def _load(self, d):
        super()._load(d)
        for i, model in enumerate(self.models):
            model._load(d["model_save_dict"][i])


class Resample1(BaseModel):
    """Sample from imbalanced dataset. Take all compounds from smaller class and
    then sample an equal number from the larger class.

    Parameters:
        model (BaseModel): Model to use for classification.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.Resample1(
        oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000)
    )
    model.fit(train['Drug'], train['Y'])
    preds = model.predict(test['Drug'])
    ------------------------------

    Note: may only be used on binary classification data.
    """

    @log_arguments
    def __init__(self, model: BaseModel, log=True):
        self.model = model
        super().__init__(log=False)

    def _fit(self, X_train, y_train):
        self.model.fit(X_train, y_train)

    def fit(self, X_train, y_train):
        y_train = np.array(y_train)
        assert (
            len(np.unique(y_train)) == 2
        ), "Resample1 can only be used on binary classification tasks"
        v1 = np.unique(y_train)[0]
        v2 = np.unique(y_train)[1]
        ix1 = np.where(y_train == v1)[0]
        ix2 = np.where(y_train == v2)[0]
        if len(ix1) > len(ix2):
            ix = np.concatenate(
                (ix2, np.random.choice(ix1, size=len(ix2), replace=False))
            )
        else:
            ix = np.concatenate(
                (ix1, np.random.choice(ix2, size=len(ix1), replace=False))
            )
        if isinstance(X_train, list):
            X_train = [X_train[i] for i in ix]
        elif isinstance(X_train, pd.Series) or isinstance(X_train, pd.DataFrame):
            X_train = X_train.iloc[ix]
        else:
            X_train = X_train[ix]
        print(len(X_train))
        super().fit(X_train, y_train[ix])

    def _predict(self, X):
        return self.model.predict(X)

    def _save(self):
        d = {"model_save": self.model._save()}
        return d

    def _load(self, d):
        self.model._load(d["model_save"])


class Resample2(BaseModel):
    """Sample from imbalanced dataset. Take all compounds from smaller class and
    then sample an equal number from the larger class."""

    @log_arguments
    def __init__(self, model: BaseModel, log=True):
        self.model = model
        super().__init__(log=False)

    def _fit(self, X_train, y_train):
        self.model.fit(X_train, y_train)

    def fit(self, X_train, y_train):
        y_train = np.array(y_train)
        assert (
            len(np.unique(y_train)) == 2
        ), "Resample2 can only be used on binary classification tasks"
        v1 = np.unique(y_train)[0]
        v2 = np.unique(y_train)[1]
        ix1 = np.where(y_train == v1)[0]
        ix2 = np.where(y_train == v2)[0]
        print(len(ix1))
        print(len(ix2))
        if len(ix1) > len(ix2):
            ix = np.concatenate(
                (
                    ix1,
                    ix2,
                    np.random.choice(ix2, size=len(ix1) - len(ix2), replace=True),
                )
            )
        else:
            ix = np.concatenate(
                (
                    ix1,
                    ix2,
                    np.random.choice(ix1, size=len(ix2) - len(ix1), replace=True),
                )
            )
        if isinstance(X_train, list):
            X_train = [X_train[i] for i in ix]
        elif isinstance(X_train, pd.Series) or isinstance(X_train, pd.DataFrame):
            X_train = X_train.iloc[ix]
        else:
            X_train = X_train[ix]
        print(len(ix))
        print(len(X_train))
        super().fit(X_train, y_train[ix])

    def _predict(self, X):
        return self.model.predict(X)

    def _save(self):
        d = {"model_save": self.model._save()}
        return d

    def _load(self, d):
        self.model._load(d["model_save"])


class ResampleAdaboost(BaseBoosting):
    """ResampleAdaBoost performs the Adaboost with sampling weighting being done
    via resampling of the dataset to create an ensembled model.

    Parameters:
        models (List[BaseModel]): list of models to use for the learners to be stacked together.
        n (int, optional): Number of times to repeat the given models. Defaults to 1.
        size (int, optional): Size of the resampled dataset. Defaults to None.
        factor (int, optional): Factor by which to resample the dataset. Defaults to 8.
        equation (str, optional): Equation to use for resampling. Defaults to "abs".
        log (bool, optional): Whether or not to log the arguments of this constructor. Defaults to True.

    Example
    ------------------------------
    import olorenchemengine as oce

    model = oce.ResampleAdaboost(
        models = [
            oce.RandomForestModel(representation = oce.MorganVecRepresentation(radius=2, nbits=2048), n_estimators = 1000),
            oce.SupportVectorMachine(representation = oce.Mol2Vec())],
        factor = 8,
        equation = 'abs'
    )
    model.fit(train['Drug'], train['Y'])
    model.predict(test['Drug'])
    ------------------------------
    """

    @log_arguments
    def __init__(
        self,
        models: List[BaseModel],
        n: int = 1,
        factor: int = 8,
        size: int = None,
        equation: str = "abs",
        log: bool = True,
        **kwargs
    ):
        """_summary_

        Args:
            models (List[BaseModel]): List of models to use in gradient boosting.
            n (int, optional): The number of times to replicate the inputted models. Defaults to 1.
            factor (int, optional): Factor to increase the number of samples resampled by. Mutually exclusive with size. Defaults to 8.
            size (int, optional): _description_. Amount of samples to use in resampling. Mutually exclusive with size. Defaults to None.
            equation (str, optional): What equation to use as loss for resampling weighting, can be "abs", "sqr", or "exp. Defaults to "abs".
            log (bool, optional): Log arguments or not. Should only be true if it is not nested. Defaults to True.
        """
        super().__init__(models, n=n, log=False, **kwargs)
        self.size = size
        self.equation = equation
        self.factor = factor

    def _fit(self, X_train, y_train):

        if self.size is None:
            size = int(len(X_train) * self.factor)
        else:
            size = self.size
        X_train = pd.DataFrame(X_train)
        y_train_df = pd.DataFrame(y_train)
        y_train = np.array(y_train)

        w = np.array([1 / len(X_train)] * len(X_train))
        self.betas = list()

        i = 0
        for model in self.models:
            i += 1

            p = w / np.sum(w)
            idxs = np.random.choice(len(X_train), size=size, replace=True, p=p)
            X_train_temp = X_train.iloc[idxs, :]
            y_train_temp = y_train_df.iloc[idxs, :].to_numpy().flatten()
            model.fit(X_train_temp, y_train_temp)
            y_pred = np.squeeze(np.array(model.predict(X_train)).flatten())
            res = np.abs(y_train - y_pred)

            D = np.max(res)
            if self.equation == "abs":
                loss = res / D
            elif self.equation == "sqr":
                loss = res**2 / D**2
            else:
                loss = 1 - np.exp(-res / D)

            loss = loss / np.max(loss)
            avg_loss = np.sum(loss * p)
            beta = avg_loss / (1 - avg_loss)

            w = w * np.power(beta, 1 - loss)
            self.betas.append(beta)

        self.betas = np.array(self.betas)

    def _predict(self, X):
        ys = list()
        for model in self.models:
            ys.append(model.predict(X))
        ys = np.array(ys)

        predictions = list()

        predictions = np.dot(ys.transpose(), np.log10(1 / self.betas)) / np.sum(
            np.log10(1 / self.betas)
        )

        return np.array(predictions)

    def _save(self):
        d = super()._save()
        if not hasattr(self, "betas"):
            return d
        d.update({"betas": self.betas})
        return d

    def _load(self, d):
        super()._load(d)
        if "betas" in d:
            self.betas = d["betas"]



from .ChemProp.main import ChemPropModel
from .GINNetwork.main import GINModel
from .HondaSTRep.main import HondaSTRep
from .HondaSTRep.operations import WordVocab
from .mol2vec.main import Mol2Vec
from .MolCLR.main import MolCLR, MolCLRVecRep
from .piCalculax import calc_pI
from .SPGNN.main import SPGNN, SPGNNVecRep
from .stoned import *



""" Module containing classes describing available attributes for visualizations.

These attributes are used to standardize the parameters of visualizations to make
them more accessible (to the extent possible)."""

import os
from abc import ABC, abstractmethod

from olorenchemengine.base_class import *
from olorenchemengine.dataset import *
from olorenchemengine.internal import *
from olorenchemengine.interpret import *
from olorenchemengine.representations import *
from olorenchemengine.uncertainty import *


class BaseAttribute(ABC):
    """Base class for all attributes for a visualization

    Parameters:
        name (str): The name of the attribute
    """

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def to_json(self) -> list:
        """
        Convert the attribute to a json object

        Parameters:
            name (str): The name of the attribute

        Returns:
            The json representation of the object
        """
        pass


class SimpleAttribute(BaseAttribute):
    """A simple attribute for a visualization is
    and attribute whose only parameter is a name.

    This describes all attributes except for the InputThreshold class,
    which requires a minimum and maximum value alongside a name."""

    def to_json(self) -> list:
        return [self.name, self.attribute_name]


class ModelSelector(SimpleAttribute):
    """Class to create a model selector attribute

    Parameters:
        name (str): The name of the attribute
    """

    attribute_name = "modelSelector"


class DatasetSelector(SimpleAttribute):
    """Class to create a dataset selector attribute

    Parameters:
        name (str): The name of the attribute
    """

    attribute_name = "datasetSelector"


class InputNumber(SimpleAttribute):
    """Class to create an input number attribute

    Parameters:
        name (str): The name of the attribute
    """

    attribute_name = "inputNumber"


class InputString(SimpleAttribute):
    """Class to create an input string attribute

    Parameters:
        name (str): The name of the attribute
    """

    attribute_name = "inputString"


class ColorPicker(SimpleAttribute):
    """Class to create a color picker attribute

    Parameters:
        name (str): The name of the attribute
    """

    attribute_name = "colorPicker"


class InputThreshold(BaseAttribute):
    """Class to create an input threshold attribute

    Parameters:
        name (str): The name of the attribute
        min (int): The minimum value of the threshold
        max (int): The maximum value of the threshold
    """

    @log_arguments
    def __init__(
        self,
        name: str,
        min: int = 0,
        max: int = 100,
    ):
        super().__init__(name)
        self.min = min
        self.max = max

    def to_json(self):
        inputThreshold = [
            self.name,
            "threshold",
            "min: " + str(self.min),
            "max: " + str(self.max),
        ]
        return inputThreshold


class AttributeSection(BaseAttribute):
    """Class to create an attribute section

    Parameters:
        attribute_name (str): The name of the attribute section
        attributes (list): The list of attributes of type BaseAttribute
    """

    @log_arguments
    def __init__(
        self,
        attribute_name: str,
        attributes: List[BaseAttribute],
    ):
        super().__init__(attribute_name)
        self.attributes = attributes

    def to_json(self):
        attribute_section = {
            self.name: [attribute.to_json() for attribute in self.attributes]
        }

        return attribute_section
